"""Various metrics that can be used to evaluate rankings.

All metrics take two parameters: `y_true` and `y_pred`. Both of these
are (n_instances, n_objects) shaped arrays of integers. We call these
arrays rankings. The element (i, j) of a ranking specifies the rank of
the jth object in the ith instance. `y_true` should be set to the
"ground truth" to evaluate against, while `y_pred` is the prediction
that should be evaluated.

Examples
--------
Lets assume we have two instances: ABCD and abcd. The "ground truth"
rankings are A > D > C > B and d < c < a < b.

We applied some ranking algorithm, which gave rankings of A > C > D > B
and d < a < b < c respectively.

Let's use some of the metrics defined here to evaluate the performance
of our ranker:

First encode the ground truth as a list of rankings. 0 is the highest
rank:
>>> y_true = [
...     [0, 3, 2, 1], # A > D > C > B, 0 is the highest rank
...     [2, 3, 1, 0], # d < c < a < b
... ]

Now similarly encode our prediction:
>>> y_pred = [
...     [0, 3, 1, 2], # A > C > D > B
...     [1, 2, 3, 0], # d < a < b < c
... ]

Evaluate with a simple zero-one loss:
>>> from keras import backend as K
>>> K.eval(zero_one_rank_loss(y_true, y_pred))
0.25

This is what we would expect: 25% of the objects were ranked at exactly the
right place. This might not be the most realistic metric, so let's try the
expected reciprocal rank instead:

>>> K.eval(err(y_true, y_pred))
0.6365559895833333
"""
from functools import partial

from keras import backend as K
import numpy as np
import tensorflow as tf

from csrank.tensorflow_util import get_instances_objects
from csrank.tensorflow_util import scores_to_rankings
from csrank.tensorflow_util import tensorify

__all__ = [
    "zero_one_rank_loss",
    "zero_one_rank_loss_for_scores",
    "zero_one_rank_loss_for_scores_ties",
    "make_ndcg_at_k_loss",
    "kendalls_tau_for_scores",
    "spearman_correlation_for_scores",
    "zero_one_accuracy",
    "zero_one_accuracy_for_scores",
    "topk_categorical_accuracy",
]


def zero_one_rank_loss(y_true, y_pred):
    y_true, y_pred = tensorify(y_true), tensorify(y_pred)
    mask = K.greater(y_true[:, None] - y_true[:, :, None], 0)
    # Count the number of mistakes (here position difference less than 0)
    mask2 = K.less(y_pred[:, None] - y_pred[:, :, None], 0)
    mask3 = K.equal(y_pred[:, None] - y_pred[:, :, None], 0)

    # Calculate Transpositions
    transpositions = tf.logical_and(mask, mask2)
    transpositions = K.sum(K.cast(transpositions, dtype="float32"), axis=[1, 2])

    n_objects = K.max(y_true) + 1
    transpositions += (
        K.sum(K.cast(mask3, dtype="float32"), axis=[1, 2]) - n_objects
    ) / 4.0
    denominator = K.cast((n_objects * (n_objects - 1.0)) / 2.0, dtype="float32")
    result = transpositions / denominator
    return K.mean(result)


def zero_one_accuracy(y_true, y_pred):
    y_true, y_pred = tensorify(y_true), tensorify(y_pred)
    n_instances, n_objects = get_instances_objects(y_true)
    equal_ranks = K.cast(K.all(K.equal(y_pred, y_true), axis=1), dtype="float32")
    denominator = K.cast(n_instances, dtype="float32")
    zero_one_loss = K.sum(equal_ranks) / denominator
    return zero_one_loss


def zero_one_rank_loss_for_scores(y_true, s_pred):
    return zero_one_rank_loss_for_scores_ties(y_true, s_pred)


def zero_one_rank_loss_for_scores_ties(y_true, s_pred):
    y_true, s_pred = tensorify(y_true), tensorify(s_pred)
    n_objects = K.cast(K.max(y_true) + 1, dtype="float32")
    mask = K.greater(y_true[:, None] - y_true[:, :, None], 0)
    mask2 = K.greater(s_pred[:, None] - s_pred[:, :, None], 0)
    mask3 = K.equal(s_pred[:, None] - s_pred[:, :, None], 0)

    # Calculate Transpositions
    transpositions = tf.logical_and(mask, mask2)
    transpositions = K.sum(K.cast(transpositions, dtype="float32"), axis=[1, 2])
    transpositions += (
        K.sum(K.cast(mask3, dtype="float32"), axis=[1, 2]) - n_objects
    ) / 4.0

    denominator = n_objects * (n_objects - 1.0) / 2.0
    result = transpositions / denominator
    return K.mean(result)


def make_ndcg_at_k_loss(k=5):
    r"""Computes the Normalized Discounted Cumulative Gain

    The Discounted Cumulative Gain is the sum of the document's relevancies,
    logarithmically discounted by their rank. That means the DCG is higher when
    the more relevant documents are highly ranked, lower otherwise.

    Concretely:

    .. math::
        \mathrm{DCG}_p = \sum_{i = 1}^p \frac{\mathit{rel}_i}{\log_2(i + 1)}

    Where :math:`\mathit{rel}_i` is the relevance of the document that is
    ranked at :math:`i`. Since this library deals with ranks, not relevances,
    it is necessary to define a conversion between the two. We define the
    relevance of an item as :math:`2^{\mathit{inv}}` where :math:`\mathit{inv}`
    is the negative rank normalized to :math:`[0, 1]`. An alternative way to
    view this is that :math:`\mathit{inv}` is the relevancy and our definition
    of ndcg exponentially discounts relevancies.

    To make the DCG comparable across different rankings (particularly rankings
    of different length), it is normalized by the ideal DCG. The resulting nDCG
    can be described as

    .. math::
        \mathrm{nDCG}_p = \frac{\mathrm{DCG}_p}{\mathrm{IDCG}_p}

    with

    .. math::
        \mathrm{IDCG}_p
        = \sum_{i = 1}^{\lvert \mathit{REL}_p \rvert} \frac{\mathit{rel}}{\log_2(i + 1)}

    where :math:`\mathit{REL}_p` is the list of relevant documents and
    :math:`\mathit{rel_i}` are the document relevancies in decreasing order.

    It follows that the nDCG is always a value in :math:`(0, 1]`, with
    :math:`1` being the best value.

    Parameters
    ----------
    k: int
        The length of the ranking for evaluation purposes. If the actual
        ranking is longer than `k`, only the (true) top `k` entries are
        considered. This is often more useful than considering the full
        ranking, for example when only a subset of the elements will actually
        be presented to a user.
    """

    def ndcg(y_true, y_pred):
        y_true, y_pred = tensorify(y_true), tensorify(y_pred)

        max_rank = K.max(y_true)

        def rank_to_relevance(rank):
            # Convert a rank to a relevance, which is (somewhat arbitrarily)
            # defined to be inversely proportional to the rank and normalized
            # to [0, 1]. Other conversion functions are possible.
            normalized_inverse = (max_rank - rank) / max_rank
            # define the relevance as 2**a
            return K.pow(2.0, normalized_inverse) - 1.0

        relevance_true = rank_to_relevance(y_true)
        relevance_pred = rank_to_relevance(y_pred)

        # Calculate ideal dcg:
        most_relevant_items, most_relevant_idx = tf.math.top_k(relevance_true, k)
        # arange starts at 0, but ranks start at 1 and the log term starts at 2
        log_term = K.log(K.arange(k, dtype="float32") + 2.0)
        # keras only natively supports the natural logarithm, have to switch base
        log2_term = log_term / K.log(2.0)
        idcg = K.sum(most_relevant_items / log2_term, axis=-1, keepdims=True)

        # Calculate actual dcg:

        # The index of the row of every element in toppred_ind, i.e.
        # [[0, 0],
        #  [1, 1]]
        row_ind = K.cumsum(K.ones_like(most_relevant_idx, dtype="int32"), axis=0) - 1

        # Indices of the k truly most relevant items, sorted by relevance. We
        # want to sort the predictions based on those indices, since that is
        # what we're trying to match.
        full_indices = K.stack([row_ind, most_relevant_idx], axis=-1)

        # Predicted relevances for the items that *should* have the top k
        # slots, ordered by the relevance rank they *should* have (so the
        # log2_term from the true predictions still has the right oder)
        top_k_preds = tf.gather_nd(relevance_pred, full_indices)

        weighted = top_k_preds / log2_term
        dcg = K.sum(weighted, axis=-1, keepdims=True)

        gain = dcg / idcg
        return gain

    return ndcg


def kendalls_tau_for_scores(y_true, y_pred):
    return 1.0 - 2.0 * zero_one_rank_loss_for_scores(y_true, y_pred)


def spearman_correlation_for_scores(y_true, y_pred):
    y_true, y_pred = tensorify(y_true), tensorify(y_pred)
    n_instances, n_objects = get_instances_objects(y_true)
    predicted_rankings = scores_to_rankings(n_objects, y_pred)
    y_true = K.cast(y_true, dtype="float32")
    sum_of_squared_distances = tf.constant(0.0)
    for i in np.arange(K.int_shape(y_pred)[1]):
        objects_pred = predicted_rankings[:, i]
        objects_true = y_true[:, i]
        t = (objects_pred - objects_true) ** 2
        sum_of_squared_distances = sum_of_squared_distances + tf.reduce_sum(t)
    denominator = K.cast(
        n_objects * (n_objects ** 2 - 1) * n_instances, dtype="float32"
    )
    spearman_correlation = 1 - (6 * sum_of_squared_distances) / denominator
    return spearman_correlation


def zero_one_accuracy_for_scores(y_true, y_pred):
    y_true, y_pred = tensorify(y_true), tensorify(y_pred)
    n_instances, n_objects = get_instances_objects(y_true)
    predicted_rankings = scores_to_rankings(n_objects, y_pred)
    y_true = K.cast(y_true, dtype="float32")
    equal_ranks = K.cast(
        K.all(K.equal(predicted_rankings, y_true), axis=1), dtype="float32"
    )
    denominator = K.cast(n_instances, dtype="float32")
    zero_one_loss = K.sum(equal_ranks) / denominator
    return zero_one_loss


def topk_categorical_accuracy(k=5):
    def topk_acc(y_true, y_pred):
        y_true, y_pred = tensorify(y_true), tensorify(y_pred)
        acc = tf.nn.in_top_k(y_pred, tf.argmax(y_true, axis=-1), k=k)
        acc = K.cast(acc, dtype="float32")
        return acc

    return topk_acc


def relevance_gain(grading, max_grade):
    """Maps a ranking (0 to `max_grade`, lower is better) to its gain.

    The gain is defined similar to the Discounted Cumulative Gain (DCG)
    metric. The value is always in [0, 1]. Therefore, it can be
    interpreted as a probability.

    Parameters
    ----------
    max_grade: float
        The highest achievable grade.
    grading: float
        A grading. Higher gradings are assumed to be better.
        0 <= grading <= max_grade must always hold.

    Tests
    -----
    >>> y_true = [0, 3, 2, 1] # (A > D > C > B)
    >>> K.eval(relevance_gain(y_true, max_grade=3)).tolist()
    [0.875, 0.0, 0.125, 0.375]
    """
    grading = tensorify(grading)
    inverse_grading = -grading + tf.cast(max_grade, grading.dtype)
    return (2 ** inverse_grading - 1) / (2 ** tf.cast(max_grade, tf.float32))


def err(y_true, y_pred, utility_function=None, probability_mapping=None):
    """Computes the Expected Reciprocal Rank or any Cascade Metric.

    ERR[1] is the cascade metric with the reciprocal rank as the utility
    function.

    Parameters
    ----------
    y_true: list of int
        The "ground truth" ranking. In this case, this does not need to
        actually be a ranking. It can be any 2 dimensional array whose
        elements can be transformed to probabilities by the
        `probability_mapping`.
    y_pred: list of int
        The predicted ranking that is to be evaluated.
    probability_mapping: list of int -> list of float
        A function that maps the elements of `y_true` to probabilities.
        Those values are then interpreted as the probability that the
        corresponding object satisfies the user's need. If `None` is
        specified, the `relevance_gain` function with `max_grade` set to
        the highest grade occurring in the grading is used.
    utility_function: int -> float
        A function that maps a rank (0 being the highest) to its
        "utility". If `None` is specified, this is defined as the
        reciprocal of the rank (resulting in the ERR metric). If a
        different utility is specified, this function can compute any
        cascade metric. Corresponds to the function represented by
        :math:`\\phi` in [1]. This will usually be a monotonically
        decreasing function, since the user is more likely to examine
        the first few results and therefore more likely to derive
        utility from them.

    Examples
    --------

    First, let's keep the default values and evaluate a ranking:
    >>> y_true = [
    ...     [0, 3, 2, 1],
    ...     [2, 3, 1, 0],
    ... ]
    >>> y_pred = [
    ...     [0, 3, 1, 2],
    ...     [1, 2, 3, 0],
    ... ]
    >>> K.eval(err(y_true, y_pred))
    0.6365559895833333

    Instead of relying on the relevance gain, we can also explicitly specify
    our own probabilities:
    >>> y_true = [
    ...     [0.3, 0.6, 0.05, 0.05],
    ...     [0.1, 0.1, 0.1, 0.7],
    ... ]
    >>> y_pred = [
    ...     [0, 3, 1, 2],
    ...     [1, 2, 3, 0],
    ... ]

    Now `y_true[i, j]` is the probability that object `j` in instance `i`
    satisfies the user's need. To use this probabilities unchanged, we need to
    override the probability mapping with the identity:

    >>> probability_mapping = lambda x: x

    Let us further specify that the rank utilities decrease in an exponential
    manner, e.g. every rank is only half as "valuable" as its predecessor:
    >>> utility_function = lambda r: 1/2**(r - 1) # start with 2**0 = 1

    We can now evaluate the metric:
    >>> K.eval(err(
    ...     y_true,
    ...     y_pred,
    ...     probability_mapping=probability_mapping,
    ...     utility_function=utility_function,
    ... ))
    0.3543499991945922

    The resulting metric is technically no longer an expected reciprocal rank,
    since the utility is not given by the reciprocal of the rank. It is a
    different version of a cascade metric. The original paper [1] called it an
    abandonment cascade (with gamma = 1/2), so let us define a new name for it:

    >>> from functools import partial
    >>> abandonment_cascade_half = partial(
    ...     err,
    ...     probability_mapping=probability_mapping,
    ...     utility_function=utility_function,
    ... )
    >>> K.eval(abandonment_cascade_half(y_true, y_pred))
    0.3543499991945922

    References
    ----------
        [1] Chapelle, Olivier, et al. "Expected reciprocal rank for graded
        relevance." Proceedings of the 18th ACM conference on Information and
        knowledge management. ACM, 2009. http://olivier.chapelle.cc/pub/err.pdf
    """
    if probability_mapping is None:
        max_grade = tf.reduce_max(y_true)
        probability_mapping = partial(relevance_gain, max_grade=max_grade)
    if utility_function is None:

        def reciprocal_rank(rank):
            return 1 / rank

        utility_function = reciprocal_rank
    y_true, y_pred = tensorify(y_true), tensorify(y_pred)

    ninstances = tf.shape(y_pred)[0]
    nobjects = tf.shape(y_pred)[1]

    # Using y_true and the probability mapping, we can derive the
    # probability that each object satisfies the users need (we need to
    # map over the flattened array and then restore the shape):
    satisfied_probs = tf.reshape(
        tf.map_fn(probability_mapping, tf.reshape(y_true, (-1,))), tf.shape(y_true)
    )

    # sort satisfied probabilities according to the predicted ranking
    rows = tf.range(0, ninstances)
    rows_cast = tf.broadcast_to(tf.reshape(rows, (-1, 1)), tf.shape(y_pred))
    full_indices = tf.stack([rows_cast, tf.cast(y_pred, tf.int32)], axis=2)
    satisfied_at_rank = tf.gather_nd(satisfied_probs, full_indices)

    not_satisfied_n_times = tf.cumprod(1 - satisfied_at_rank, axis=1, exclusive=True)

    # And from the positions predicted in y_pred we can further derive
    # the utilities of each object given their position:
    utilities = tf.map_fn(
        utility_function, tf.range(1, nobjects + 1), dtype=tf.float64,
    )

    discount_at_rank = tf.cast(not_satisfied_n_times, tf.float64) * tf.reshape(
        utilities, (1, -1)
    )
    discounted_document_values = (
        tf.cast(satisfied_at_rank, tf.float64) * discount_at_rank
    )
    results = tf.reduce_sum(discounted_document_values, axis=1)

    return K.mean(results)



import logging
import os

import numpy as np

from csrank.constants import OBJECT_RANKING
from .util import sub_sampling_rankings
from ..dataset_reader import DatasetReader

try:
    import h5py
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("h5py", "data")


class SentenceOrderingDatasetReader(DatasetReader):
    def __init__(self, n_dims=25, n_objects=5, **kwargs):
        super(SentenceOrderingDatasetReader, self).__init__(
            learning_problem=OBJECT_RANKING,
            dataset_folder="sentence_ordering",
            **kwargs,
        )
        self.logger = logging.getLogger(name=SentenceOrderingDatasetReader.__name__)
        dimensions = {25, 50, 100, 200}
        d_files = ["test_{}_dim.h5", "train_{}_dim.h5"]
        self.n_objects = n_objects
        if n_dims not in dimensions:
            raise ValueError(f"n_dims must be one of {dimensions}")
        self.logger.info("Loading the dataset with {} features".format(n_dims))
        self.train_file = os.path.join(self.dirname, d_files[1]).format(n_dims)
        self.test_file = os.path.join(self.dirname, d_files[0]).format(n_dims)

        self.__load_dataset__()

    def __load_dataset__(self):
        file = h5py.File(self.train_file, "r")
        self.X_train, self.Y_train = self.get_rankings_dict(file)
        file = h5py.File(self.test_file, "r")
        self.X_test, self.Y_test = self.get_rankings_dict(file)
        self.logger.info("Done loading the dataset")

    def get_rankings_dict(self, file):
        lengths = file["lengths"]
        X = dict()
        Y = dict()
        for ranking_length in np.array(lengths):
            self.X = np.array(file["X_{}".format(ranking_length)])
            self.Y = np.array(file["Y_{}".format(ranking_length)])
            X[ranking_length], Y[ranking_length] = self.X, self.Y
            self.__check_dataset_validity__()
        return X, Y

    def sub_sampling_for_dictionary(self):
        X = []
        Y = []
        for n in self.X_train.keys():
            if n > self.n_objects:
                x, y = sub_sampling_rankings(
                    self.X_train[n], self.Y_train[n], n_objects=self.n_objects
                )
                if len(X) == 0:
                    X = np.copy(x)
                    Y = np.copy(y)
                else:
                    X = np.concatenate([X, x], axis=0)
                    Y = np.concatenate([Y, y], axis=0)
        if self.n_objects in self.X_train:
            X = np.concatenate([X, np.copy(self.X_train[self.n_objects])], axis=0)
            Y = np.concatenate([Y, np.copy(self.Y_train[self.n_objects])], axis=0)
        self.logger.info(
            "Sampled instances {} objects {}".format(X.shape[0], X.shape[1])
        )
        return X, Y

    def splitter(self, iter):
        pass

    def get_train_test_datasets(self, n_datasets):
        pass

    def get_dataset_dictionaries(self):
        return self.X_train, self.Y_train, self.X_test, self.Y_test

    def get_single_train_test_split(self):
        self.X, self.Y = self.sub_sampling_for_dictionary()
        self.__check_dataset_validity__()
        return self.X, self.Y, self.X_test, self.Y_test



import math

from keras.losses import binary_crossentropy
import numpy as np
from sklearn.utils import check_random_state
import tensorflow as tf

from csrank.learner import Learner
from csrank.numpy_util import sigmoid
from csrank.util import print_dictionary
from csrank.util import progress_bar


class FATELinearCore(Learner):
    def __init__(
        self,
        n_hidden_set_units=32,
        learning_rate=1e-3,
        batch_size=256,
        loss_function=binary_crossentropy,
        epochs_drop=300,
        drop=0.1,
        random_state=None,
        **kwargs,
    ):
        self.n_hidden_set_units = n_hidden_set_units
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.random_state = random_state
        self.loss_function = loss_function
        self.epochs_drop = epochs_drop
        self.drop = drop
        self.current_lr = None
        self.weight1 = None
        self.bias1 = None
        self.weight2 = None
        self.bias2 = None
        self.optimizer = None

    def _construct_model_(self, n_objects):
        self.X = tf.placeholder(
            "float32", [None, n_objects, self.n_object_features_fit_]
        )
        self.Y = tf.placeholder("float32", [None, n_objects])
        std = 1 / np.sqrt(self.n_object_features_fit_)
        self.b1 = tf.Variable(
            self.random_state_.normal(loc=0, scale=std, size=self.n_hidden_set_units),
            dtype=tf.float32,
        )
        self.W1 = tf.Variable(
            self.random_state_.normal(
                loc=0,
                scale=std,
                size=(self.n_object_features_fit_, self.n_hidden_set_units),
            ),
            dtype=tf.float32,
        )
        self.W2 = tf.Variable(
            self.random_state_.normal(
                loc=0,
                scale=std,
                size=(self.n_object_features_fit_ + self.n_hidden_set_units),
            ),
            dtype=tf.float32,
        )
        self.b2 = tf.Variable(
            self.random_state_.normal(loc=0, scale=std, size=1), dtype=tf.float32
        )

        set_rep = (
            tf.reduce_mean(tf.tensordot(self.X, self.W1, axes=1), axis=1) + self.b1
        )

        self.set_rep = tf.reshape(
            tf.tile(set_rep, tf.constant([1, n_objects])),
            (-1, n_objects, self.n_hidden_set_units),
        )
        self.X_con = tf.concat([self.X, self.set_rep], axis=-1)
        scores = tf.sigmoid(tf.tensordot(self.X_con, self.W2, axes=1) + self.b2)
        scores = tf.cast(scores, tf.float32)
        self.loss = self.loss_function(self.Y, scores)
        self.optimizer = tf.train.GradientDescentOptimizer(self.learning_rate).minimize(
            self.loss
        )

    def step_decay(self, epoch):
        step = math.floor((1 + epoch) / self.epochs_drop)
        self.current_lr = self.learning_rate * math.pow(self.drop, step)
        self.optimizer = tf.train.GradientDescentOptimizer(self.current_lr).minimize(
            self.loss
        )

    def fit(
        self, X, Y, epochs=10, callbacks=None, validation_split=0.1, verbose=0, **kwd
    ):
        self.random_state_ = check_random_state(self.random_state)
        # Global Variables Initializer
        n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        self._construct_model_(self.n_objects_fit_)
        init = tf.global_variables_initializer()

        with tf.Session() as tf_session:
            tf_session.run(init)
            self._fit_(X, Y, epochs, n_instances, tf_session, verbose)
            training_cost = tf_session.run(self.loss, feed_dict={self.X: X, self.Y: Y})
            self.logger.info(
                "Fitting completed {} epochs done with loss {}".format(
                    epochs, training_cost.mean()
                )
            )
            self.weight1 = tf_session.run(self.W1)
            self.bias1 = tf_session.run(self.b1)
            self.weight2 = tf_session.run(self.W2)
            self.bias2 = tf_session.run(self.b2)

    def _fit_(self, X, Y, epochs, n_instances, tf_session, verbose):
        try:
            for epoch in range(epochs):
                for start in range(0, n_instances, self.batch_size):
                    end = np.min([start + self.batch_size, n_instances])
                    tf_session.run(
                        self.optimizer,
                        feed_dict={self.X: X[start:end], self.Y: Y[start:end]},
                    )
                    if verbose == 1:
                        progress_bar(end, n_instances, status="Fitting")
                if verbose == 1:
                    c = tf_session.run(self.loss, feed_dict={self.X: X, self.Y: Y})
                    print("Epoch {}: cost {} ".format((epoch + 1), np.mean(c)))
                if (epoch + 1) % 100 == 0:
                    c = tf_session.run(self.loss, feed_dict={self.X: X, self.Y: Y})
                    self.logger.info(
                        "Epoch {}: cost {} ".format((epoch + 1), np.mean(c))
                    )
                self.step_decay(epoch)
        except KeyboardInterrupt:
            self.logger.info("Interrupted")
            c = tf_session.run(self.loss, feed_dict={self.X: X, self.Y: Y})
            self.logger.info("Epoch {}: cost {} ".format((epoch + 1), np.mean(c)))

    def _predict_scores_fixed(self, X, **kwargs):
        n_instances, n_objects, n_features = X.shape
        assert n_features == self.n_object_features_fit_
        rep = np.mean(np.dot(X, self.weight1), axis=1) + self.bias1
        rep = np.tile(rep[:, np.newaxis, :], (1, n_objects, 1))
        X_n = np.concatenate((X, rep), axis=2)
        scores = np.dot(X_n, self.weight2) + self.bias2
        scores = sigmoid(scores)
        return scores

    def set_tunable_parameters(
        self,
        n_hidden_set_units=32,
        learning_rate=1e-3,
        batch_size=128,
        epochs_drop=300,
        drop=0.1,
        **point,
    ):
        """
            Set tunable parameters of the FETA-network to the values provided.

            Parameters
            ----------
            n_hidden_set_units: int
                Number of hidden units in each layer of the scoring network
            learning_rate: float
                Learning rate of the stochastic gradient descent algorithm used by the network
            batch_size: int
                Batch size to use during training
            epochs_drop: int
                The epochs after which the learning rate is decreased
            drop: float
                The percentage with which the learning rate is decreased
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        self.n_hidden_set_units = n_hidden_set_units
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.epochs_drop = epochs_drop
        self.drop = drop
        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support tunable parameters"
                " called: {}".format(print_dictionary(point))
            )



import itertools

from keras import backend as K
import numpy as np
from numpy.testing import assert_almost_equal
import pytest
from pytest import approx

from csrank.metrics import err
from csrank.metrics import kendalls_tau_for_scores
from csrank.metrics import make_ndcg_at_k_loss
from csrank.metrics import spearman_correlation_for_scores
from csrank.metrics import zero_one_accuracy
from csrank.metrics import zero_one_accuracy_for_scores
from csrank.metrics import zero_one_rank_loss
from csrank.metrics import zero_one_rank_loss_for_scores
from csrank.metrics import zero_one_rank_loss_for_scores_ties
from csrank.metrics_np import err_np
from csrank.metrics_np import kendalls_tau_for_scores_np
from csrank.metrics_np import spearman_correlation_for_scores_np
from csrank.metrics_np import spearman_correlation_for_scores_scipy
from csrank.metrics_np import zero_one_accuracy_for_scores_np
from csrank.metrics_np import zero_one_rank_loss_for_scores_np
from csrank.metrics_np import zero_one_rank_loss_for_scores_ties_np
from csrank.numpy_util import ranking_ordering_conversion


@pytest.fixture(scope="module", params=[(False), (True)], ids=["NoTies", "Ties"])
def problem_for_pred(request):
    ties = request.param
    y_true = np.arange(5)[None, :]
    # We test the error by swapping one adjacent pair:
    if ties:
        y_pred = np.array([[0, 2, 1, 2, 3]])
    else:
        y_pred = np.array([[0, 2, 1, 3, 4]])
    return y_true, y_pred, ties


@pytest.fixture(scope="module", params=[(False), (True)], ids=["NoTies", "Ties"])
def problem_for_scores(request):
    ties = request.param
    y_true = np.arange(5)[None, :]
    # We test the error by swapping one adjacent pair:
    if ties:
        y_scores = np.array([[1.0, 0.8, 0.9, 0.8, 0.6]])
    else:
        y_scores = np.array([[1.0, 0.8, 0.9, 0.7, 0.6]])
    return y_true, y_scores, ties


def test_zero_one_rank_loss(problem_for_pred):
    y_true, y_pred, ties = problem_for_pred
    score = zero_one_rank_loss(y_true, y_pred)
    real_score = K.eval(score)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.15]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.1]))


def test_zero_one_rank_loss_for_scores(problem_for_scores):
    y_true, y_scores, ties = problem_for_scores
    score = zero_one_rank_loss_for_scores(y_true, y_scores)
    real_score = K.eval(score)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.15]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.1]))

    score = zero_one_rank_loss_for_scores_ties(y_true, y_scores)
    real_score = K.eval(score)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.15]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.1]))

    y_true, y_scores, ties = problem_for_scores
    real_score = zero_one_rank_loss_for_scores_np(y_true, y_scores)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.15]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.1]))

    real_score = zero_one_rank_loss_for_scores_ties_np(y_true, y_scores)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.15]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.1]))


def test_zero_one_accuracy(problem_for_pred):
    y_true, y_pred, ties = problem_for_pred

    score = zero_one_accuracy(y_true, y_pred)
    real_score = K.eval(score)
    assert_almost_equal(actual=real_score, desired=np.array([0.0]))

    y_true, y_pred, ties = problem_for_pred

    real_score = zero_one_accuracy_for_scores_np(y_true, y_pred)
    assert_almost_equal(actual=real_score, desired=np.array([0.0]))


def test_ndcg(problem_for_pred):
    # ties don't matter here because it doesn't change the two highest predictions
    y_true, y_pred, _ties = problem_for_pred
    # We have:
    # y_true = [0, 1, 2, 3, 4]
    # y_pred = [0, 2, 1, 2, 3]

    # Inverted (with max_rank = 4) that is
    # y_true_inv = [4, 3, 2, 1, 0]
    # y_pred_inv = [4, 2, 3, 2, 1]

    # And normalized to [0, 1] this gives us the relevance:
    # rel_true = [1, 3/4, 1/2, 1/4, 0]
    # rel_pred = [1, 1/2, 3/4, 1/2, 1/4]

    # With this we can first compute the ideal dcg, considering only the first
    # k=2 elements (all logs are base 2, equality is approximate):
    idcg = (2 ** 1 - 1) / np.log2(2) + (2 ** (3 / 4) - 1) / np.log2(3)  # = 1.43

    # And the dcg of the predictions at the same positions as the elements we
    # considered for the idcg (i.e. the "true" best elements):
    dcg = (2 ** 1 - 1) / np.log2(2) + (2 ** (1 / 2) - 1) / np.log2(3)  # = 1.26

    # Now the gain is:
    expected_gain = dcg / idcg  # = 0.882

    ndcg = make_ndcg_at_k_loss(k=2)
    real_gain = K.eval(ndcg(y_true, y_pred))

    assert_almost_equal(actual=real_gain, desired=expected_gain, decimal=5)


def test_kendalls_tau_for_scores(problem_for_scores):
    y_true, y_pred, ties = problem_for_scores

    score = kendalls_tau_for_scores(y_true, y_pred)
    real_score = K.eval(score)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.7]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.8]))

    real_score = kendalls_tau_for_scores_np(y_true, y_pred)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.7]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.8]))


def test_spearman_for_scores(problem_for_scores):
    y_true_tensor, y_scores_tensor, ties = problem_for_scores

    score = spearman_correlation_for_scores(y_true_tensor, y_scores_tensor)
    real_score = K.eval(score)
    if ties:
        # We do not handle ties for now
        assert True
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.9]))

    y_true, y_scores, ties = problem_for_scores

    real_score = spearman_correlation_for_scores_scipy(y_true, y_scores)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.8207827]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.9]))

    real_score = spearman_correlation_for_scores_np(y_true, y_scores)
    if ties:
        assert True
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.9]))


def test_zero_one_accuracy_for_scores(problem_for_scores):
    y_true_tensor, y_scores_tensor, ties = problem_for_scores

    score = zero_one_accuracy_for_scores(y_true_tensor, y_scores_tensor)
    real_score = K.eval(score)
    if ties:
        assert_almost_equal(actual=real_score, desired=np.array([0.0]))
    else:
        assert_almost_equal(actual=real_score, desired=np.array([0.0]))


def test_err_perfect_first_trumps_many_good():
    """Tests that a perfect document at rank 1 trumps later rankings.

    The authors of [1] list this as a motivating example. A ranking that
    puts a "perfect" document at rank 1 (i.e. one that is almost certain
    to satisfy the user's needs) should trump one that puts a "good" one
    at rank 1, regardless of the documents at later ranks. The reasoning
    is that later ranks won't need to be examined when the first is
    already sufficient.

    References
    ----------
        [1] Chapelle, Olivier, et al. "Expected reciprocal rank for graded
        relevance." Proceedings of the 18th ACM conference on Information and
        knowledge management. ACM, 2009. http://olivier.chapelle.cc/pub/err.pdf
    """
    y_true = ranking_ordering_conversion([range(20)])

    # gets the "perfect" one right, everything else wrong
    perfect_first = ranking_ordering_conversion(
        [[0, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1]]
    )

    # does pretty good for most, but ranks the "perfect" one wrong
    all_good = ranking_ordering_conversion(
        [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 0]]
    )

    assert K.eval(err(y_true, perfect_first)) > K.eval(err(y_true, all_good))


def test_err_against_manually_verified_example():
    """Compares the implementation against a manual calculation."""
    y_true = ranking_ordering_conversion([[1, 2, 0]])
    y_pred = ranking_ordering_conversion([[2, 1, 0]])
    # The resulting probabilities that each document satisfies the
    # user's need:
    # [2**1-1, 2**2-1, 2**0 - 1] / 2**2 = [1/4, 3/4, 0]
    # Multiplied by the respective rank utilities (1/(r+1)):
    # [(1/4)/3, (3/4)/2, 0/1] = [1/12, 3/8, 0]
    # The resulting ERR:

    # We ranked object 2 first, which has a true rank of 1 and therefore
    # (with the relevance gain probability mapping) a probability of
    # (2**(2-1)-1) / 2**2 = 1/4
    # of matching the user's need. It is at rank 0, which has utility
    # 1/(0+1) = 1.

    # Object 1 is next. True rank of 0, probability
    # (2**(2-0)-1) / 2**2 = 3/4
    # and utility
    # 1/(1+1) = 1/2.

    # Object 0 last. True rank of 2, probability
    # (2**(2-2)-1) / 2**2 = 0
    # and utility
    # 1/(2+1) = 1/3.

    # The resulting expected utility:
    # 1/4 * 1 + (1 - 1/4) * 3/4 * 1/2 + (1 - 1/4) * (1 - 3/4) * 0 * 1/3
    # = 17/32
    # Approx because comparing floats is inherently error-prone.
    assert K.eval(err(y_true, y_pred)) == approx(17 / 32)


def test_err_implementations_equivalent():
    """Spot-checks equivalence of plain python and tf implementations"""
    # A simple grading where each grade occurs once. We want to check
    # for equivalence at every permutation of this grading.
    elems = np.array([4, 3, 2, 1, 0])
    y_true = np.reshape(elems, (1, -1))
    # Spot check some permutations (5! / 20 = 6 checks are performed)
    for perm in list(itertools.permutations(elems))[::20]:
        perm = np.reshape(perm, (1, -1))
        assert K.eval(err(y_true, perm)) == approx(err_np(y_true, perm))



import copy
import inspect
import os

import numpy as np
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import StandardScaler

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")

# Get hotel_dataset for each city in a dictionary
POPINDEX = "popindex"
MUENCHEN = "muenchen"
HAMBURG = "hamburg"
FRANKFURT = "frankfurt"
DUSSLEDORF = "dussledorf"
BERLIN = "berlin"
DATASET_FOLDER = "hotel_dataset"

__all__ = ["TripAdviosrDataset"]


class TripAdviosrDataset(object):
    def __init__(self, **kwargs):
        self.city_names = [BERLIN, DUSSLEDORF, MUENCHEN, HAMBURG, FRANKFURT]
        self.MIN_HOTEL_RANK_LENGTH = 110
        dirname = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        mypath = os.path.join(dirname, DATASET_FOLDER, "trip_advisor.csv")
        full_dataset = pd.read_csv(mypath)
        index_1 = full_dataset.columns.get_loc(BERLIN)
        city_indices = np.arange(index_1, index_1 + 5)
        self.datasets_dictionaries = dict()
        self.datasets_dictionaries[BERLIN] = (
            full_dataset[full_dataset.berlin == 1]
            .drop(full_dataset.columns[city_indices], axis=1)
            .sort_values(by=[POPINDEX], ascending=[True])
        )
        self.datasets_dictionaries[DUSSLEDORF] = (
            full_dataset[full_dataset.duesseldorf == 1]
            .drop(full_dataset.columns[city_indices], axis=1)
            .sort_values(by=[POPINDEX], ascending=[True])
        )
        self.datasets_dictionaries[FRANKFURT] = (
            full_dataset[full_dataset.frankfurt == 1]
            .drop(full_dataset.columns[city_indices], axis=1)
            .sort_values(by=[POPINDEX], ascending=[True])
        )
        self.datasets_dictionaries[HAMBURG] = (
            full_dataset[full_dataset.hamburg == 1]
            .drop(full_dataset.columns[city_indices], axis=1)
            .sort_values(by=[POPINDEX], ascending=[True])
        )
        self.datasets_dictionaries[MUENCHEN] = (
            full_dataset[full_dataset.muenchen == 1]
            .drop(full_dataset.columns[city_indices], axis=1)
            .sort_values(by=[POPINDEX], ascending=[True])
        )
        for key, value in self.datasets_dictionaries.items():
            popindex = value[POPINDEX]
            value.drop(POPINDEX, axis=1, inplace=True)
            value.insert(len(list(value.columns)), POPINDEX, popindex)
            value.reset_index(drop=True)
        self.all_headers = self.datasets_dictionaries[BERLIN].columns[0:-1]
        self.all_dataframes = pd.concat(
            list(self.datasets_dictionaries.values()), ignore_index=True
        )

    # Get features and rankings for given hotel_dataset for a city
    def data_for_city(self, city, features_to_remove=[]):

        if city not in self.city_names:
            raise ValueError(
                f"Data is not present for the city, hotel_dataset is present only for: {set(self.city_names)}"
            )
        dataframe = copy.deepcopy(self.datasets_dictionaries[city])
        dataframe.drop(features_to_remove, axis=1, inplace=True)
        headers = list(dataframe.columns[0:-1])
        # index_to_remove = [np.where(headers == f)[0] for f in features_to_remove]
        array = dataframe.as_matrix()
        features = array[:, 0:-1]
        # get actual and normalized ranks
        actual_ranks = array[:, -1]
        scaler = MinMaxScaler()
        normalized_ranks = np.array(actual_ranks)
        normalized_ranks = scaler.fit_transform(normalized_ranks.reshape(-1, 1)).T[0]
        # get features and filtered features
        features = np.array(
            [np.log(np.array(features[:, i]) + 1) for i in range(len(features[0]))]
        )
        features = np.array(features.T)
        scaler = StandardScaler()
        features = scaler.fit_transform(features)
        actual_ranks = np.arange(len(actual_ranks))

        features = features[None, :]
        actual_ranks = actual_ranks[None, :]
        normalized_ranks = normalized_ranks[None, :]

        return headers, features, actual_ranks, normalized_ranks

    def data_for_cities(self, cities=[BERLIN], features_to_remove=[]):
        X = []
        Y = []
        norm_ranks = []
        for city in cities:
            headers, features, actual_ranks, normalized_ranks = self.data_for_city(
                city, features_to_remove
            )
            X.append(np.array(features[0]))
            Y.append(np.array(actual_ranks[0]))
            norm_ranks.append((np.array(normalized_ranks[0])))
        X = np.array(X)
        Y = np.array(Y)
        norm_ranks = np.array(norm_ranks)
        return X, Y, norm_ranks, headers



from .cmpnet_core import CmpNetCore
from .fate_linear import FATELinearCore
from .fate_network import FATENetwork
from .fate_network import FATENetworkCore
from .feta_linear import FETALinearCore
from .feta_network import FETANetwork
from .pairwise_svm import PairwiseSVM
from .ranknet_core import RankNetCore

__all__ = [
    "CmpNetCore",
    "FATELinearCore",
    "FATENetwork",
    "FATENetworkCore",
    "FETALinearCore",
    "FETANetwork",
    "PairwiseSVM",
    "RankNetCore",
]



import logging

from sklearn.utils import check_random_state

from csrank.constants import DISCRETE_CHOICE
from csrank.dataset_reader.util import standardize_features
from .util import sub_sampling_discrete_choices_from_relevance
from ..letor_ranking_dataset_reader import LetorRankingDatasetReader


class LetorRankingDiscreteChoiceDatasetReader(LetorRankingDatasetReader):
    def __init__(self, random_state=None, n_objects=5, **kwargs):
        super(LetorRankingDiscreteChoiceDatasetReader, self).__init__(
            learning_problem=DISCRETE_CHOICE, **kwargs
        )
        self.logger = logging.getLogger(
            LetorRankingDiscreteChoiceDatasetReader.__name__
        )
        self.random_state = check_random_state(random_state)
        self.n_objects = n_objects
        self.__load_dataset__()

    def sub_sampling_function(self, X, Y):
        return sub_sampling_discrete_choices_from_relevance(
            Xt=X, Yt=Y, n_objects=self.n_objects
        )

    def splitter(self, iter):
        pass

    def get_train_test_datasets(self, n_datasets):
        pass

    def get_single_train_test_split(self):
        X_train, Y_train = self.sub_sampling_from_dictionary(train_test="train")
        X_test, Y_test = self.sub_sampling_from_dictionary(train_test="test")
        X_train, X_test = standardize_features(X_train, X_test)
        return X_train, Y_train, X_test, Y_test



from keras import backend as K
import tensorflow as tf

from csrank.tensorflow_util import tensorify

__all__ = [
    "hinged_rank_loss",
    "make_smooth_ndcg_loss",
    "smooth_rank_loss",
    "plackett_luce_loss",
]


def identifiable(loss_function):
    def wrap_loss(y_true, y_pred):
        alpha = 1e-4
        ss = tf.reduce_sum(tf.square(y_pred), axis=1)
        return alpha * ss + loss_function(y_true, y_pred)

    return wrap_loss


@identifiable
def hinged_rank_loss(y_true, y_pred):
    y_true, y_pred = tensorify(y_true), tensorify(y_pred)
    mask = K.cast(K.greater(y_true[:, None] - y_true[:, :, None], 0), dtype="float32")
    diff = y_pred[:, :, None] - y_pred[:, None]
    hinge = K.maximum(mask * (1 - diff), 0)
    n = K.sum(mask, axis=(1, 2))
    return K.sum(hinge, axis=(1, 2)) / n


@identifiable
def smooth_rank_loss(y_true, y_pred):
    y_true, y_pred = tensorify(y_true), tensorify(y_pred)
    mask = K.cast(K.greater(y_true[:, None] - y_true[:, :, None], 0), dtype="float32")
    exped = K.exp(y_pred[:, None] - y_pred[:, :, None])
    result = K.sum(exped * mask, axis=[1, 2])
    return result / K.sum(mask, axis=(1, 2))


@identifiable
def plackett_luce_loss(y_true, s_pred):
    y_true = tf.cast(y_true, dtype="int32")
    s_pred = tf.cast(s_pred, dtype="float32")
    m = tf.shape(y_true)[1]
    raw_max = tf.reduce_max(s_pred, axis=1, keepdims=True)
    max_elem = tf.stop_gradient(
        tf.where(tf.is_finite(raw_max), raw_max, tf.zeros_like(raw_max))
    )
    exped = tf.exp(tf.subtract(s_pred, max_elem))
    masks = tf.greater_equal(y_true, tf.range(m)[:, None, None])
    tri = exped * tf.cast(masks, tf.float32)
    lse = tf.reduce_sum(tf.log(tf.reduce_sum(tri, axis=2)), axis=0)
    return lse - tf.reduce_sum(s_pred, axis=1)


def make_smooth_ndcg_loss(y_true, y_pred):
    y_true, y_pred = tensorify(y_true), tensorify(y_pred)
    n_objects = K.max(y_true) + 1.0
    y_true_f = K.cast(y_true, "float32")
    relevance = n_objects - y_true_f - 1.0
    log_term = K.log(relevance + 2.0) / K.log(2.0)
    exp_relevance = K.pow(2.0, relevance) - 1.0
    gains = exp_relevance / log_term

    # Calculate ideal dcg:
    idcg = K.sum(gains, axis=-1)

    # Calculate smoothed dcg:
    exped = K.exp(y_pred)
    exped = exped / K.sum(exped, axis=-1, keepdims=True)
    # toppred, toppred_ind = tf.nn.top_k(gains * exped, k)
    return 1 - K.sum(exped * gains, axis=-1) / idcg



from .baseline import RandomBaselineDC
from .cmpnet_discrete_choice import CmpNetDiscreteChoiceFunction
from .fate_discrete_choice import FATEDiscreteChoiceFunction
from .fatelinear_discrete_choice import FATELinearDiscreteChoiceFunction
from .feta_discrete_choice import FETADiscreteChoiceFunction
from .fetalinear_discrete_choice import FETALinearDiscreteChoiceFunction
from .generalized_nested_logit import GeneralizedNestedLogitModel
from .mixed_logit_model import MixedLogitModel
from .model_selector import ModelSelector
from .multinomial_logit_model import MultinomialLogitModel
from .nested_logit_model import NestedLogitModel
from .paired_combinatorial_logit import PairedCombinatorialLogit
from .pairwise_discrete_choice import PairwiseSVMDiscreteChoiceFunction
from .ranknet_discrete_choice import RankNetDiscreteChoiceFunction

__all__ = [
    "RandomBaselineDC",
    "CmpNetDiscreteChoiceFunction",
    "FATEDiscreteChoiceFunction",
    "FATELinearDiscreteChoiceFunction",
    "FETADiscreteChoiceFunction",
    "FETALinearDiscreteChoiceFunction",
    "GeneralizedNestedLogitModel",
    "MixedLogitModel",
    "ModelSelector",
    "MultinomialLogitModel",
    "NestedLogitModel",
    "PairedCombinatorialLogit",
    "PairwiseSVMDiscreteChoiceFunction",
    "RankNetDiscreteChoiceFunction",
]



from abc import ABCMeta
from itertools import product
import logging
import pickle as pk

import numpy as np

try:
    import pymc3 as pm
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pymc3", "probabilistic")


class ModelSelector(metaclass=ABCMeta):
    def __init__(
        self,
        learner_cls,
        parameter_keys,
        model_params,
        fit_params,
        model_path,
        **kwargs,
    ):
        self.priors = [
            [pm.Normal, {"mu": 0, "sd": 10}],
            [pm.Laplace, {"mu": 0, "b": 10}],
        ]
        self.uniform_prior = [pm.Uniform, {"lower": -20, "upper": 20}]
        self.prior_indices = np.arange(len(self.priors))
        self.parameter_f = [
            (pm.Normal, {"mu": 0, "sd": 5}),
            (pm.Cauchy, {"alpha": 0, "beta": 1}),
            0,
            -5,
            5,
        ]
        self.parameter_s = [
            (pm.HalfCauchy, {"beta": 1}),
            (pm.HalfNormal, {"sd": 0.5}),
            (pm.Exponential, {"lam": 0.5}),
            (pm.Uniform, {"lower": 1, "upper": 10}),
            10,
        ]
        # ,(pm.HalfCauchy, {'beta': 2}), (pm.HalfNormal, {'sd': 1}),(pm.Exponential, {'lam': 1.0})]
        self.learner_cls = learner_cls
        self.model_params = model_params
        self.fit_params = fit_params
        self.parameter_keys = parameter_keys
        self.parameters = list(product(self.parameter_f, self.parameter_s))
        pf_arange = np.arange(len(self.parameter_f))
        ps_arange = np.arange(len(self.parameter_s))
        self.parameter_ind = list(product(pf_arange, ps_arange))
        self.model_path = model_path
        self.models = dict()
        self.logger = logging.getLogger(ModelSelector.__name__)

    def fit(self, X, Y):
        model_args = dict()
        for param_key in self.parameter_keys:
            model_args[param_key] = self.uniform_prior
        self.logger.info("Uniform Prior")
        self.model_params["model_args"] = model_args
        key = "{}_uniform_prior".format(self.parameter_keys)
        self.fit_learner(X, Y, key)
        for j, param in enumerate(self.parameters):
            self.logger.info("mu: {}, sd/b: {}".format(*self.parameter_ind[j]))
            if len(self.parameter_keys) == 2:
                for i1, i2 in product(self.prior_indices, self.prior_indices):
                    prior1 = self.priors[i1]
                    prior2 = self.priors[i2]
                    self.logger.info("Priors {}, {}".format(i1, i2))
                    model_args = dict()
                    k1 = list(prior1[1].keys())
                    k2 = list(prior2[1].keys())
                    prior1[1] = dict(zip(k1, param))
                    prior2[1] = dict(zip(k2, param))
                    model_args[self.parameter_keys[0]] = prior1
                    model_args[self.parameter_keys[1]] = prior2
                    key = "{}_{}_{}_{}_mu_{}_sd_{}".format(
                        self.parameter_keys[0],
                        i1,
                        self.parameter_keys[1],
                        i2,
                        self.parameter_ind[j][0],
                        self.parameter_ind[j][1],
                    )
                    self.model_params["model_args"] = model_args
                    self.fit_learner(X, Y, key)
            else:
                for i, prior in enumerate(self.priors):
                    self.logger.info("Prior {}".format(i))
                    model_args = dict()
                    k1 = list(prior[1].keys())
                    prior[1] = dict(zip(k1, param))
                    model_args[self.parameter_keys[0]] = prior
                    self.model_params["model_args"] = model_args
                    key = "{}_{}_mu_{}_sd_{}".format(
                        self.parameter_keys[0],
                        i,
                        self.parameter_ind[j][0],
                        self.parameter_ind[j][1],
                    )
                    self.fit_learner(X, Y, key)

    def fit_learner(self, X, Y, key):
        learner = self.learner_cls(**self.model_params)
        try:
            learner.fit(X, Y, **self.fit_params)
            self.models[key] = {"model": learner.model, "trace": learner.trace}
            self.logger.info("Model done for priors key {}".format(key))
            f = open(self.model_path, "wb")
            pk.dump(self.models, f)
            f.close()
        except Exception as e:
            self.logger.error("Error for parameters {}: {}".format(key, e))



from itertools import product
import logging

import numpy as np
from sklearn.utils import check_random_state

from csrank.learner import Learner
import csrank.numpy_util as npu
import csrank.theano_util as ttu
from csrank.util import print_dictionary
from .discrete_choice import DiscreteObjectChooser
from .likelihoods import create_weight_dictionary
from .likelihoods import fit_pymc3_model
from .likelihoods import likelihood_dict
from .likelihoods import LogLikelihood

try:
    import pymc3 as pm
    from pymc3.variational.callbacks import CheckParametersConvergence
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pymc3", "probabilistic")

try:
    import theano
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")


class GeneralizedNestedLogitModel(DiscreteObjectChooser, Learner):
    def __init__(
        self,
        n_nests=None,
        loss_function="None",
        regularization="l2",
        alpha=5e-2,
        random_state=None,
        **kwd,
    ):
        """
            Create an instance of the Generalized Nested Logit model for learning the discrete choice function. This
            model divides objects into subsets called nests, such that the each object is associtated to each nest to some degree.
            This model structure is 1-layer of hierarchy and the :math:`\\lambda` for each nest :math:`B_k` signifies the degree of independence
            and  :math:`1-\\lambda` signifies the correlations between the object in it. We learn two weight vectors and the  :math:`\\lambda s`.
            The probability of choosing an object :math:`x_i` from the given query set :math:`Q` is defined by product
            of choosing the nest in which :math:`x_i` exists and then choosing the the object from the nest.

            .. math::

                P(x_i \\lvert Q) = P_i = \\sum_{\\substack{B_k \\in \\mathcal{B} \\ i \\in B_k}}P_{i \\lvert B_k} P_{B_k} \\enspace ,


            The discrete choice for the given query set :math:`Q` is defined as:

            .. math::

                dc(Q) := \\operatorname{argmax}_{x_i \\in Q }  \\; P(x_i \\lvert Q)

            Parameters
            ----------
            n_objects: int
                Number of objects in each query set
            n_nests : int range : [2,n_objects/2]
                The number of nests/subsets in which the objects are divided.
                This may not surpass half the amount of objects this model will
                be trained on.
            loss_function : string , {‘categorical_crossentropy’, ‘binary_crossentropy’, ’categorical_hinge’}
                Loss function to be used for the discrete choice decision from the query set
            regularization : string, {‘l1’, ‘l2’}, string
               Regularizer function (L1 or L2) applied to the `kernel` weights matrix
            alpha: float (range : [0,1])
                The lower bound of the correlations between the objects in a nest
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the algorithms

            References
            ----------
                [1] Kenneth E Train. „Discrete choice methods with simulation“. In: Cambridge university press, 2009. Chap GEV, pp. 87–111.

                [2] Kenneth Train. Qualitative choice analysis. Cambridge, MA: MIT Press, 1986

                [3] Chieh-Hua Wen and Frank S Koppelman. „The generalized nested logit model“. In: Transportation Research Part B: Methodological 35.7 (2001), pp. 627–641

        """
        self.logger = logging.getLogger(GeneralizedNestedLogitModel.__name__)

        self.n_nests = n_nests
        self.alpha = alpha
        self.loss_function = likelihood_dict.get(loss_function, None)

        self.random_state = random_state
        known_regularization_functions = {"l1", "l2"}
        if regularization not in known_regularization_functions:
            raise ValueError(
                f"Regularization function {regularization} is unknown. Must be one of {known_regularization_functions}"
            )
        self.regularization = regularization
        self._config = None
        self.model = None
        self.trace = None
        self.trace_vi = None
        self.Xt = None
        self.Yt = None
        self.p = None
        self._config = None
        self.threshold = 43e5

    @property
    def model_configuration(self):
        """
            Constructs the dictionary containing the priors for the weight vectors for the model according to the
            regularization function. The parameters are:
                * **weights** : Weights to evaluates the utility of the objects
                * **weights_k** : Weights to evaluates the fractional allocation of each object in :math:'Q' to each nest

            For ``l1`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{b}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Laplace}(\\text{mu}=\\text{mu}_w, \\text{b}=\\text{b}_w)

            For ``l2`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{sd}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Normal}(\\text{mu}=\\text{mu}_w, \\text{sd}=\\text{sd}_w)

            Returns
            -------
                configuration : dict
                    Dictionary containing the priors applies on the weights
        """
        if self._config is None:
            if self.regularization == "l2":
                weight = pm.Normal
                prior = "sd"
            elif self.regularization == "l1":
                weight = pm.Laplace
                prior = "b"
            self._config = {
                "weights": [
                    weight,
                    {
                        "mu": (pm.Normal, {"mu": 0, "sd": 5}),
                        prior: (pm.HalfCauchy, {"beta": 1}),
                    },
                ],
                "weights_ik": [
                    weight,
                    {
                        "mu": (pm.Normal, {"mu": 0, "sd": 5}),
                        prior: (pm.HalfCauchy, {"beta": 1}),
                    },
                ],
            }
            self.logger.info(
                "Creating model with config {}".format(print_dictionary(self._config))
            )
        return self._config

    def get_probabilities(self, utility, lambda_k, alpha_ik):
        """
            This method calculates the probability of choosing an object from the query set using the following parameters of the model which are used:

                * **weights** (:math:`w`): Weights to get the utility of the object :math:`Y_i = U(x_i) = w \\cdot x_i`
                * **weights_k** (:math:`w_k`): Weights to get fractional allocation of each object :math:'x_j'  in :math:'Q' to each nest math:`B_k` as :math:`\\alpha_{ik} = w_k \\cdot x_i`.
                * **lambda_k** (:math:`\\lambda_k`): Lambda for nest :math:`B_k` for correlations between the obejcts.

            The probability of choosing the object :math:`x_i` from the query set :math:`Q`:

            .. math::
                P_i = \\sum_{\\substack{B_k \\in \\mathcal{B} \\ i \\in B_k}} P_{i \\lvert {B_k}} P_{B_k} \\enspace where, \\\\
                P_{B_k} = \\frac{{\\left(\\sum_{j \\in B_k} {\\left(\\alpha_{jk} \\boldsymbol{e}^{V_j} \\right)}^ {^{1}/{\\lambda_k}} \\right)}^{\\lambda_k}}{\\sum_{\\ell = 1}^{K} {\\left( \\sum_{j \\in B_{\\ell}} {\\left( \\alpha_{j\\ell} \\boldsymbol{e}^{V_j} \\right)}^{^{1}/{\\lambda_\\ell}} \\right)^{\\lambda_{\\ell}}}} \\\\
                P_{{i} \\lvert {B_k}} = \\frac{{\\left(\\alpha_{ik} \\boldsymbol{e}^{V_i} \\right)}^{^{1}/{\\lambda_k}}}{\\sum_{j \\in B_k} {\\left(\\alpha_{jk} \\boldsymbol{e}^{V_j} \\right)}^{^{1}/{\\lambda_k}}} \\enspace ,


            Parameters
            ----------
            utility : theano tensor
                (n_instances, n_objects)
                Utility :math:`Y_i` of the objects :math:`x_i \\in Q` in the query sets
            lambda_k : theano tensor (range : [alpha, 1.0])
                (n_nests)
                Measure of independence amongst the obejcts in each nests
            alpha_ik : theano tensor
                (n_instances, n_objects, n_nests)
                Fractional allocation of each object :math:`x_i` in each nest math:`B_k`

            Returns
            -------
            p : theano tensor
                (n_instances, n_objects)
                Choice probabilities :math:`P_i` of the objects :math:`x_i \\in Q` in the query sets

        """
        n_nests = self.n_nests
        n_instances, n_objects = utility.shape
        pik = tt.zeros((n_instances, n_objects, n_nests))
        sum_per_nest = tt.zeros((n_instances, n_nests))
        for i in range(n_nests):
            uti = (utility + tt.log(alpha_ik[:, :, i])) * 1 / lambda_k[i]
            sum_n = ttu.logsumexp(uti)
            pik = tt.set_subtensor(pik[:, :, i], tt.exp(uti - sum_n))
            sum_per_nest = tt.set_subtensor(
                sum_per_nest[:, i], sum_n[:, 0] * lambda_k[i]
            )
        pnk = tt.exp(sum_per_nest - ttu.logsumexp(sum_per_nest))
        pnk = pnk[:, None, :]
        p = pik * pnk
        p = p.sum(axis=2)
        return p

    def _get_probabilities_np(self, utility, lambda_k, alpha_ik):
        n_nests = self.n_nests
        n_instances, n_objects = utility.shape
        pik = np.zeros((n_instances, n_objects, n_nests))
        sum_per_nest_x = np.zeros((n_instances, n_nests))
        for i in range(n_nests):
            uti = (utility + np.log(alpha_ik[:, :, i])) * 1 / lambda_k[i]
            sum_n = npu.logsumexp(uti)
            pik[:, :, i] = np.exp(uti - sum_n)
            sum_per_nest_x[:, i] = sum_n[:, 0] * lambda_k[i]
        pnk = np.exp(sum_per_nest_x - npu.logsumexp(sum_per_nest_x))
        pnk = pnk[:, None, :]
        p = pik * pnk
        p = p.sum(axis=2)
        return p

    def construct_model(self, X, Y):
        """
            Constructs the nested logit model by applying priors on weight vectors **weights** and **weights_k** as per
            :meth:`model_configuration`. Then we apply a uniform prior to the :math:`\\lambda s`, i.e.
            :math:`\\lambda s \\sim Uniform(\\text{alpha}, 1.0)`.The probability of choosing the object :math:`x_i` from the
            query set :math:`Q = \\{x_1, \\ldots ,x_n\\}` is evaluated in :meth:`get_probabilities`.

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Preferences in the form of discrete choices for given objects

            Returns
            -------
             model : pymc3 Model :class:`pm.Model`
        """
        self.random_state_ = check_random_state(self.random_state)
        if np.prod(X.shape) > self.threshold:
            upper_bound = int(self.threshold / np.prod(X.shape[1:]))
            indices = self.random_state_.choice(X.shape[0], upper_bound, replace=False)
            X = X[indices, :, :]
            Y = Y[indices, :]
        self.logger.info(
            "Train Set instances {} objects {} features {}".format(*X.shape)
        )
        with pm.Model() as self.model:
            self.Xt = theano.shared(X)
            self.Yt = theano.shared(Y)
            shapes = {
                "weights": self.n_object_features_fit_,
                "weights_ik": (self.n_object_features_fit_, self.n_nests),
            }
            weights_dict = create_weight_dictionary(self.model_configuration, shapes)

            alpha_ik = tt.dot(self.Xt, weights_dict["weights_ik"])
            alpha_ik = ttu.softmax(alpha_ik, axis=2)
            utility = tt.dot(self.Xt, weights_dict["weights"])
            lambda_k = pm.Uniform("lambda_k", self.alpha, 1.0, shape=self.n_nests)
            self.p = self.get_probabilities(utility, lambda_k, alpha_ik)
            LogLikelihood(
                "yl", loss_func=self.loss_function, p=self.p, observed=self.Yt
            )
        self.logger.info("Model construction completed")

    def fit(
        self,
        X,
        Y,
        sampler="variational",
        tune=500,
        draws=500,
        vi_params={
            "n": 20000,
            "method": "advi",
            "callbacks": [CheckParametersConvergence()],
        },
        **kwargs,
    ):
        """
            Fit a generalized nested logit model on the provided set of queries X and choices Y of those objects. The
            provided queries and corresponding preferences are of a fixed size (numpy arrays). For learning this network
            the categorical cross entropy loss function for each object :math:`x_i \\in Q` is defined as:

            .. math::

                C_{i} =  -y(i)\\log(P_i) \\enspace,

            where :math:`y` is ground-truth discrete choice vector of the objects in the given query set :math:`Q`.
            The value :math:`y(i) = 1` if object :math:`x_i` is chosen else :math:`y(i) = 0`.

            Parameters
            ----------
            X : numpy array (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array (n_instances, n_objects)
                Choices for given objects in the query
            sampler : {‘variational’, ‘metropolis’, ‘nuts’}, string
                The sampler used to estimate the posterior mean and mass matrix from the trace

                    * **variational** : Run inference methods to estimate posterior mean and diagonal mass matrix
                    * **metropolis** : Use the MAP as starting point and Metropolis-Hastings sampler
                    * **nuts** : Use the No-U-Turn sampler
            vi_params : dict
                The parameters for the **variational** inference method
            draws : int
                The number of samples to draw. Defaults to 500. The number of tuned samples are discarded by default
            tune : int
                Number of iterations to tune, defaults to 500. Ignored when using 'SMC'. Samplers adjust
                the step sizes, scalings or similar during tuning. Tuning samples will be drawn in addition
                to the number specified in the `draws` argument, and will be discarded unless
                `discard_tuned_samples` is set to False.
            **kwargs :
                Keyword arguments for the fit function of :meth:`pymc3.fit`or :meth:`pymc3.sample`
        """
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        if self.n_nests is None:
            # TODO this looks like a bug to me, but it was already done this way
            # before (moved out of __init__). The `n_objects` summand probably
            # should be removed.
            self.n_nests = self.n_objects_fit_ + int(self.n_objects_fit_ / 2)
        self.construct_model(X, Y)
        fit_pymc3_model(self, sampler, draws, tune, vi_params, **kwargs)

    def _predict_scores_fixed(self, X, **kwargs):
        mean_trace = dict(pm.summary(self.trace)["mean"])
        weights = np.array(
            [
                mean_trace["weights[{}]".format(i)]
                for i in range(self.n_object_features_fit_)
            ]
        )
        lambda_k = np.array(
            [mean_trace["lambda_k[{}]".format(i)] for i in range(self.n_nests)]
        )
        weights_ik = np.zeros((self.n_object_features_fit_, self.n_nests))
        for i, k in product(range(self.n_object_features_fit_), range(self.n_nests)):
            weights_ik[i][k] = mean_trace["weights_ik[{},{}]".format(i, k)]
        alpha_ik = np.dot(X, weights_ik)
        alpha_ik = npu.softmax(alpha_ik, axis=2)
        utility = np.dot(X, weights)
        p = self._get_probabilities_np(utility, lambda_k, alpha_ik)
        return p

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return DiscreteObjectChooser.predict_for_scores(self, scores, **kwargs)

    def set_tunable_parameters(
        self, alpha=None, n_nests=None, loss_function=None, regularization="l2", **point
    ):
        """
            Set tunable parameters of the Nested Logit model to the values provided.

            Parameters
            ----------
            alpha: float (range : [0,1])
                The lower bound of the correlations between the objects in a nest
            n_nests: int (range : [2,n_objects])
                The number of nests in which the objects are divided
            loss_function : string , {‘categorical_crossentropy’, ‘binary_crossentropy’, ’categorical_hinge’}
                Loss function to be used for the discrete choice decision from the query set
            regularization : string, {‘l1’, ‘l2’}, string
               Regularizer function (L1 or L2) applied to the `kernel` weights matrix
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        if alpha is not None:
            self.alpha = alpha
        # TODO see the comment for n_nests above
        if n_nests is None:
            self.n_nests = self.n_objects_fit + int(self.n_objects_fit / 2)
        else:
            self.n_nests = n_nests
        if loss_function is not None:
            if loss_function not in likelihood_dict:
                raise ValueError(
                    f"Loss function {loss_function} is unknown. Must be one of {set(likelihood_dict.keys())}"
                )
            self.loss_function = likelihood_dict.get(loss_function, None)
        self.regularization = regularization
        self.model = None
        self.trace = None
        self.trace_vi = None
        self.Xt = None
        self.Yt = None
        self.p = None
        self._config = None
        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support tunable parameters"
                " called: {}".format(print_dictionary(point))
            )



import numpy as np
from scipy.stats import rankdata


def replace_inf_np(x):
    if np.any(np.isinf(x)):
        x[np.isinf(x)] = 2e300
        x[np.isnan(x)] = 2e300
    return x


def replace_nan_np(p):
    if np.any(np.isnan(p)):
        p[np.isnan(p)] = 1.0
        p[np.isinf(p)] = 1.0
        p = normalize(p)
    return p


def logsumexp(x, axis=1):
    max_x = x.max(axis=axis, keepdims=True)
    return max_x + np.log(np.sum(np.exp(x - max_x), axis=axis, keepdims=True))


def softmax(x, axis=1):
    """
        Take softmax for the given numpy array.
        :param axis: The axis around which the softmax is applied
        :param x: array-like, shape (n_samples, ...)
        :return: softmax taken around the given axis
    """
    lse = logsumexp(x, axis=axis)
    return np.exp(x - lse)


def sigmoid(x):
    x = 1.0 / (1.0 + np.exp(-x))
    return x


def normalize(x, axis=1):
    """
        Normalize the given two dimensional numpy array around the row.
        :param axis: The axis around which the norm is applied
        :param x: theano or numpy array-like, shape (n_samples, n_objects)
        :return: normalize the array around the axis=1
    """
    return x / np.sum(x, axis=axis, keepdims=True)


def scores_to_rankings(score_matrix):
    mask3 = np.equal(score_matrix[:, None] - score_matrix[:, :, None], 0)
    n_objects = score_matrix.shape[1]
    ties = np.sum(np.sum(mask3, axis=(1, 2)) - n_objects)
    rankings = np.empty_like(score_matrix)
    if ties > 0:
        for i, s in enumerate(score_matrix):
            rankings[i] = len(s) - rankdata(s)
    else:
        orderings = np.argsort(score_matrix, axis=1)[:, ::-1]
        rankings = np.argsort(orderings, axis=1)
    return rankings


def ranking_ordering_conversion(input):
    """Converts a ranking to an ordering.

    A ranking is a list of object ranks, with the object identities
    being determined by the list indices. In contrast, an ordering is a
    list of object identities with the array indices determining the
    order.

    Examples
    --------
    Consider the ranking 0 > 3 > 1 > 2. The 1st object is at the 2nd
    position, the 2nd object on the 3rd position etc. The resulting
    ranking is:

    >>> ranking = [[0, 2, 3, 1]]

    Which can be converted back to an ordering:

    >>> ranking_ordering_conversion(ranking).tolist()
    [[0, 3, 1, 2]]

    Parameters
    ----------
    ranking: 2d array of integers, unique within each row
        The object rankings.
    """
    output = np.argsort(input, axis=1)
    return output



import logging
import math

from keras import backend as K
from keras.callbacks import Callback
import numpy as np

from csrank.tunable import Tunable
from csrank.util import print_dictionary


class EarlyStoppingWithWeights(Callback, Tunable):
    def __init__(
        self,
        monitor="val_loss",
        min_delta=0,
        patience=0,
        verbose=0,
        mode="auto",
        baseline=None,
        restore_best_weights=False,
        **kwargs,
    ):
        """Stop training when a monitored quantity has stopped improving.

        Parameters
        ----------
        monitor: string
            Quantity to be monitored, could be the loss or an accuracy value, which is monitored by the model.
            In case of accuracy the we check the change in increase, while in case of loss we check the change in
            decrease of the loss
        min_delta: float
            Minimum change in the monitored quantity to qualify as an improvement, i.e. an absolute change of less
            than min_delta, will count as no improvement
        patience: unsigned int
            number of epochs with no improvement after which training will be stopped
        verbose : bool
            verbosity mode, 1: print, 0 to not
        mode: one of {auto, min, max}.
            In `min` mode, training will stop when the quantity monitored has stopped decreasing; in `max` mode
            it will stop when the quantity monitored has stopped increasing; in `auto` mode, the direction is
            automatically inferred from the name of the monitored quantity
        baseline: float
            Baseline value for the monitored quantity to reach. Training will stop if the model doesn't show
            improvement over the baseline
        restore_best_weights: bool
            whether to restore model weights from the epoch with the best value of the monitored quantity.
            If False, the model weights obtained at the last step of training are used
        **kwargs
            Keyword arguments for the callback
        """
        super(EarlyStoppingWithWeights, self).__init__()
        self.monitor = monitor
        self.baseline = baseline
        self.patience = patience
        self.verbose = verbose
        self.min_delta = min_delta
        self.wait = 0
        self.stopped_epoch = 0
        self.restore_best_weights = restore_best_weights
        self.best_weights = None

        known_modes = {"auto", "min", "max"}
        if mode not in known_modes:
            raise ValueError(
                f"EarlyStopping mode {mode} is unknown, must be one of {known_modes}"
            )

        if mode == "min":
            self.monitor_op = np.less
        elif mode == "max":
            self.monitor_op = np.greater
        else:
            if "acc" in self.monitor:
                self.monitor_op = np.greater
            else:
                self.monitor_op = np.less

        if self.monitor_op == np.greater:
            self.min_delta *= 1
        else:
            self.min_delta *= -1
        self.logger = logging.getLogger(EarlyStoppingWithWeights.__name__)

    def on_train_begin(self, logs=None):
        # Allow instances to be re-used
        self.wait = 0
        self.stopped_epoch = 0
        if self.baseline is not None:
            self.best = self.baseline
        else:
            self.best = np.Inf if self.monitor_op == np.less else -np.Inf

    def on_epoch_end(self, epoch, logs=None):
        self.stopped_epoch += 1
        current = logs.get(self.monitor)
        self.best_weights = self.model.get_weights()
        if current is None:
            self.logger.warning(
                "Early stopping conditioned on metric `%s` which is not available. Available metrics are: %s"
                % (self.monitor, ",".join(list(logs.keys()))),
                RuntimeWarning,
            )
            return
        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.best_weights = self.model.get_weights()
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.model.stop_training = True

    def on_train_end(self, logs=None):
        if self.stopped_epoch > 0:
            self.logger.info(
                "Setting best weights for final epoch {}".format(self.stopped_epoch)
            )
            self.model.set_weights(self.best_weights)

    def set_tunable_parameters(self, patience=300, min_delta=2, **point):
        """Set tunable parameters of the EarlyStopping callback to the values provided.

        Parameters
        ----------
        patience: unsigned int
            Number of epochs with no improvement after which training will be stopped.
        min_delta: float
            Minimum change in the monitored quantity to qualify as an improvement, i.e. an absolute change of less
            than min_delta, will count as no improvement.
        point: dict
            Dictionary containing parameter values which are not tuned for the network
        """
        self.patience = patience
        self.min_delta = min_delta
        if len(point) > 0:
            self.logger.warning(
                "This callback does not support"
                " tunable parameters"
                " called: {}".format(print_dictionary(point))
            )


class LRScheduler(Callback, Tunable):
    def __init__(self, epochs_drop=300, drop=0.1, verbose=0, **kwargs):
        """Learning rate scheduler with step-decay function

        Parameters
        ----------
        epochs_drop: unsigned int
            The number of epochs after which the learning rate is reduced
        drop: float [0,1):
            The percentage of the learning rate which needs to be dropped
        verbose: bool or int in {0,1}
            int. 0: quiet, 1: update messages
        **kwargs
            Keyword arguments for the callback
        """
        super(LRScheduler, self).__init__(**kwargs)
        self.verbose = verbose
        self.epochs_drop = epochs_drop
        self.drop = drop
        self.initial_lr = None
        self.logger = logging.getLogger(LRScheduler.__name__)

    def step_decay(self, epoch):
        """The step-decay function, which takes the current epoch and update the learning rate according to the
        formulae.

        .. math::
            lr = lr_0 * d_r^{\\lfloor \\frac{e}{e_{\\text{drop}}}\\rfloor}

        where :math:`lr_0` is the learning rate at the zeroth epoch and :math:`0 < d_r < 1` is the rate with which
        the learning rate should be reduced, :math:`e` is the current epoch and :math:`e_{\\text{drop}}`
        is the number of epochs after which the learning rate is decreased.

        Parameters
        ----------
        epoch: unsigned int
            Current epoch
        """
        step = math.floor((1 + epoch) / self.epochs_drop)
        new_lr = self.initial_lr * math.pow(self.drop, step)
        return new_lr

    def on_epoch_begin(self, epoch, logs=None):
        if not hasattr(self.model.optimizer, "lr"):
            raise ValueError('Optimizer must have a "lr" attribute.')
        if epoch == 0:
            self.initial_lr = float(K.get_value(self.model.optimizer.lr))
        lr = self.step_decay(epoch)
        K.set_value(self.model.optimizer.lr, lr)
        if self.verbose > 0:
            print(
                "\nEpoch %05d: LearningRateScheduler setting learning "
                "rate to %s." % (epoch + 1, lr)
            )

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs["lr"] = K.get_value(self.model.optimizer.lr)

    def set_tunable_parameters(self, epochs_drop=300, drop=0.1, **point):
        """Set tunable parameters of the LRScheduler callback to the values provided.

        Parameters
        ----------
        epochs_drop: unsigned int
            The number of epochs after which the learning rate is reduced
        drop: float [0,1):
            The percentage of the learning rate which needs to be dropped
        point: dict
            Dictionary containing parameter values which are not tuned for the network
        """
        self.epochs_drop = epochs_drop
        self.drop = drop
        if len(point) > 0:
            self.logger.warning(
                "This callback does not support"
                " tunable parameters"
                " called: {}".format(print_dictionary(point))
            )


class DebugOutput(Callback):
    def __init__(self, delta=100, **kwargs):
        """Logging the epochs when done.

        Parameters
        ----------
        delta: unsigned int
            The number of epochs after which the message is logged
        kwargs:
            Keyword arguments
        """
        super(DebugOutput, self).__init__(**kwargs)
        self.delta = delta
        self.epoch = 0
        self.logger = logging.getLogger(DebugOutput.__name__)

    def on_train_end(self, logs=None):
        self.logger.debug("Total number of epochs: {}".format(self.epoch))

    def on_train_begin(self, logs=None):
        self.epoch = 0

    def on_epoch_end(self, epoch, logs=None):
        self.epoch += 1
        if self.epoch % self.delta == 0:
            self.logger.debug("Epoch {} of the training finished.".format(self.epoch))



from abc import ABCMeta
import logging
import os

import numpy as np

from csrank.constants import DYAD_RANKING
from .dataset_reader import DatasetReader

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")


class SushiDatasetReader(DatasetReader, metaclass=ABCMeta):
    def __init__(self, **kwargs):
        super(SushiDatasetReader, self).__init__(dataset_folder="sushi", **kwargs)
        self.object_datafile = os.path.join(self.dirname, "sushi3.txt")
        self.user_datafile = os.path.join(self.dirname, "sushi3_u.txt")
        self.rankings_small = os.path.join(self.dirname, "sushi_small.txt")
        self.rankings_big = os.path.join(self.dirname, "sushi_big.txt")
        self.__load_dataset__()
        self.__check_dataset_validity__()
        self.logger = logging.getLogger(SushiDatasetReader.__name__)

    def __load_dataset__(self):
        # Code to concatenate the context feature to each object feature
        # c = np.tile(f1[np.newaxis, :], (f2.shape[0], 1))
        # combined = np.concatenate((c, f2), axis=1)
        objects = pd.read_csv(self.object_datafile, sep="\t")
        users = pd.read_csv(self.user_datafile, sep="\t")

        small_rankings = pd.read_csv(self.rankings_small, sep=" ")
        big_rankings = pd.read_csv(self.rankings_big, sep=" ")
        rankings_a = small_rankings.as_matrix()[:, 2:13]
        rankings_b = big_rankings.as_matrix()[:, 2:13]

        feature_objects = objects.as_matrix()[:, 2:10]
        feature_users = users.as_matrix()[:, 1:12]
        feature_objects[:, 2] = [int("{0:b}".format(i)) for i in feature_objects[:, 2]]
        feature_objects = np.array(feature_objects, dtype=float)
        feature_users = np.array(feature_users, dtype=float)
        for i in range(len(feature_objects)):
            feature_objects[i][0:3] = np.log(feature_objects[i][0:3] + 2)
        feature_users = np.log(feature_users + 2)
        X = []
        rankings = []
        Xc = []
        i = 0
        for ranking1, ranking2 in zip(rankings_a, rankings_b):
            objects = feature_objects[ranking1, :]
            ranking = np.arange(len(ranking1))
            np.random.shuffle(ranking)
            objects = objects[ranking, :]
            X.append(objects)
            rankings.append(ranking)

            objects = feature_objects[ranking2, :]
            ranking = np.arange(len(ranking2))
            np.random.shuffle(ranking)
            objects = objects[ranking, :]
            X.append(objects)
            rankings.append(ranking)

            Xc.append(feature_users[i])
            i += 1

        self.X = np.array(X)
        self.Y = np.array(rankings)
        if self.learning_problem == DYAD_RANKING:
            self.Xc = np.array(Xc)
            self.logger(Xc.shape)
        else:
            self.Xc = None



from keras import backend as K
import numpy as np
import pytest
import tensorflow as tf

from csrank import SyntheticIterator
from csrank.tensorflow_util import tensorify
from csrank.tuning import check_learner_class


def test_tensorify():
    a = np.array([1.0, 2.0])
    out = tensorify(a)
    assert isinstance(out, tf.Tensor)

    b = K.zeros((5, 3))
    out = tensorify(b)
    assert b == out


def test_synthetic_iterator():
    def func(a, b):
        return (b, a)

    it = SyntheticIterator(dataset_function=func, a=41, b=2)
    for i, (x, y) in enumerate(it):
        if i == 1:
            break
        assert x == 2
        assert y == 41


def test_check_ranker_class():
    class MockClass(object):
        def __init__(self):
            pass

    ranker = MockClass()
    with pytest.raises(AttributeError):
        check_learner_class(ranker)



from keras import backend as K
import numpy as np
from numpy.testing import assert_almost_equal

from csrank.losses import hinged_rank_loss
from csrank.losses import plackett_luce_loss
from csrank.losses import smooth_rank_loss

decimal = 3


def test_hinged_rank_loss():
    y_true = np.arange(5)[None, :]
    y_true_tensor = K.constant(y_true)

    # Predicting all 0, gives an error of 1.0:
    assert_almost_equal(
        actual=K.eval(
            hinged_rank_loss(
                y_true_tensor, K.constant(np.array([[0.0, 0.0, 0.0, 0.0, 0.0]]))
            )
        ),
        desired=np.array([1.0]),
        decimal=decimal,
    )

    # Predicting the correct ranking improves, but penalizes by difference of
    # scores:
    assert_almost_equal(
        actual=K.eval(
            hinged_rank_loss(
                y_true_tensor, K.constant(np.array([[0.2, 0.1, 0.0, -0.1, -0.2]]))
            )
        ),
        desired=np.array([0.8]),
        decimal=decimal,
    )


def test_plackett_luce_loss():
    y_true = np.arange(5)[None, :]
    y_true_tensor = K.constant(y_true)
    assert_almost_equal(
        actual=K.eval(
            plackett_luce_loss(
                y_true_tensor, K.constant(np.array([[0.0, 0.0, 0.0, 0.0, 0.0]]))
            )
        ),
        desired=np.array([4.78749]),
        decimal=decimal,
    )


def test_smooth_rank_loss():
    y_true = np.arange(5)[None, :]
    y_true_tensor = K.constant(y_true)

    # Predicting all 0, gives an error of 1.0:
    assert_almost_equal(
        actual=K.eval(
            smooth_rank_loss(
                y_true_tensor, K.constant(np.array([[0.0, 0.0, 0.0, 0.0, 0.0]]))
            )
        ),
        desired=np.array([1.0]),
        decimal=decimal,
    )

    # Predicting the correct ranking improves, but penalizes by difference of
    # scores:
    assert_almost_equal(
        actual=K.eval(
            smooth_rank_loss(
                y_true_tensor, K.constant(np.array([[0.2, 0.1, 0.0, -0.1, -0.2]]))
            )
        ),
        desired=np.array([0.82275984]),
    )



import logging

import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.utils import check_random_state

from csrank.choicefunction.util import BinaryCrossEntropyLikelihood
from csrank.choicefunction.util import create_weight_dictionary
from csrank.discretechoice.likelihoods import fit_pymc3_model
from csrank.learner import Learner
import csrank.theano_util as ttu
from csrank.util import print_dictionary
from .choice_functions import ChoiceFunctions

try:
    import pymc3 as pm
    from pymc3.variational.callbacks import CheckParametersConvergence
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pymc3", "probabilistic")

try:
    import theano
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")


class GeneralizedLinearModel(ChoiceFunctions, Learner):
    def __init__(self, regularization="l2", random_state=None, **kwargs):
        """
            Create an instance of the GeneralizedLinearModel model for learning the choice function. This model is
            adapted from the multinomial logit model :class:`csrank.discretechoice.multinomial_logit_model.MultinomialLogitModel`.
            The utility score for each object in query set :math:`Q` is defined as :math:`U(x) = w \\cdot x`,
            where :math:`w` is the weight vector. The probability of choosing an object :math:`x_i` is defined by taking
            sigmoid over the utility scores:

            .. math::

                P(x_i \\lvert Q) = \\frac{1}{1+exp(-U(x_i))}

            The choice set is defined as:

            .. math::

                c(Q) = \\{ x_i \\in Q \\lvert \\, P(x_i \\lvert Q) > t \\}

            Parameters
            ----------
            regularization : string, optional
                Regularization technique to be used for estimating the weights
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the algorithms

            References
            ----------
                [1] Kenneth E Train. „Discrete choice methods with simulation“. In: Cambridge university press, 2009. Chap Logit, pp. 41–86.

                [2] Kenneth Train. Qualitative choice analysis. Cambridge, MA: MIT Press, 1986
        """
        self.logger = logging.getLogger(GeneralizedLinearModel.__name__)
        known_regularization_functions = {"l1", "l2"}
        if regularization not in known_regularization_functions:
            raise ValueError(
                f"Regularization function {regularization} is unknown. Must be one of {known_regularization_functions}"
            )
        self.regularization = regularization
        self.random_state = random_state
        self.model = None
        self.trace = None
        self.trace_vi = None
        self.Xt = None
        self.Yt = None
        self.p = None
        self.threshold = 0.5

    @property
    def model_configuration(self):
        """
            Constructs the dictionary containing the priors for the weight vectors for the model according to the
            regularization function. The parameters are:
                * **weights** : Weights to evaluates the utility of the objects

            For ``l1`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{b}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Laplace}(\\text{mu}=\\text{mu}_w, \\text{b}=\\text{b}_w)

            For ``l2`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{sd}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Normal}(\\text{mu}=\\text{mu}_w, \\text{sd}=\\text{sd}_w)
        """
        if self.regularization == "l2":
            weight = pm.Normal
            prior = "sd"
        elif self.regularization == "l1":
            weight = pm.Laplace
            prior = "b"
        configuration = {
            "weights": [
                weight,
                {
                    "mu": (pm.Normal, {"mu": 0, "sd": 10}),
                    prior: (pm.HalfCauchy, {"beta": 1}),
                },
            ]
        }
        self.logger.info(
            "Creating default config {}".format(print_dictionary(configuration))
        )
        return configuration

    def construct_model(self, X, Y):
        """
            Constructs the linear logit model which evaluated the utility score as :math:`U(x) = w \\cdot x`, where
            :math:`w` is the weight vector. The probability of choosing the object :math:`x_i` from the query set
            :math:`Q = \\{x_1, \\ldots ,x_n\\}` is:

            .. math::

                P_i =  P(x_i \\lvert Q) = \\frac{1}{1+exp(-U(x_i))}

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Preferences in form of Choices for given objects

            Returns
            -------
             model : pymc3 Model :class:`pm.Model`
        """
        self.logger.info(
            "Creating model_args config {}".format(
                print_dictionary(self.model_configuration)
            )
        )
        with pm.Model() as self.model:
            self.Xt = theano.shared(X)
            self.Yt = theano.shared(Y)
            shapes = {"weights": self.n_object_features_fit_}
            # shapes = {'weights': (self.n_object_features_fit_, 3)}
            weights_dict = create_weight_dictionary(self.model_configuration, shapes)
            intercept = pm.Normal("intercept", mu=0, sd=10)
            utility = tt.dot(self.Xt, weights_dict["weights"]) + intercept
            self.p = ttu.sigmoid(utility)
            BinaryCrossEntropyLikelihood("yl", p=self.p, observed=self.Yt)
        self.logger.info("Model construction completed")

    def fit(
        self,
        X,
        Y,
        sampler="variational",
        tune=500,
        draws=500,
        tune_size=0.1,
        thin_thresholds=1,
        vi_params={
            "n": 20000,
            "method": "advi",
            "callbacks": [CheckParametersConvergence()],
        },
        verbose=0,
        **kwargs,
    ):
        """
            Fit a generalized logit model on the provided set of queries X and choices Y of those objects. The
            provided queries and corresponding preferences are of a fixed size (numpy arrays). For learning this network
            the binary cross entropy loss function for each object :math:`x_i \\in Q` is defined as:

            .. math::

                C_{i} =  -y(i)\\log(P_i) - (1 - y(i))\\log(1 - P_i) \\enspace,

            where :math:`y` is ground-truth choice vector of the objects in the given query set :math:`Q`.
            The value :math:`y(i) = 1` if object :math:`x_i` is chosen else :math:`y(i) = 0`.

            Parameters
            ----------
            X : numpy array (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array (n_instances, n_objects)
                Choices for given objects in the query
            sampler : {‘variational’, ‘metropolis’, ‘nuts’}, string
                The sampler used to estimate the posterior mean and mass matrix from the trace

                    * **variational** : Run inference methods to estimate posterior mean and diagonal mass matrix
                    * **metropolis** : Use the MAP as starting point and Metropolis-Hastings sampler
                    * **nuts** : Use the No-U-Turn sampler
            vi_params : dict
                The parameters for the **variational** inference method
            draws : int
                The number of samples to draw. Defaults to 500. The number of tuned samples are discarded by default
            tune : int
                Number of iterations to tune, defaults to 500. Ignored when using 'SMC'. Samplers adjust
                the step sizes, scalings or similar during tuning. Tuning samples will be drawn in addition
                to the number specified in the `draws` argument, and will be discarded unless
                `discard_tuned_samples` is set to False.
            tune_size: float (range : [0,1])
                Percentage of instances to split off to tune the threshold for the choice function
            thin_thresholds: int
                The number of instances of scores to skip while tuning the threshold
            verbose : bool
                Print verbose information
            **kwargs :
                Keyword arguments for the fit function
        """
        self.random_state_ = check_random_state(self.random_state)
        if tune_size > 0:
            X_train, X_val, Y_train, Y_val = train_test_split(
                X, Y, test_size=tune_size, random_state=self.random_state_
            )
            try:
                self._fit(
                    X_train, Y_train, sampler=sampler, vi_params=vi_params, **kwargs
                )
            finally:
                self.logger.info(
                    "Fitting utility function finished. Start tuning threshold."
                )
                self.threshold = self._tune_threshold(
                    X_val, Y_val, thin_thresholds=thin_thresholds, verbose=verbose
                )
        else:
            self._fit(
                X,
                Y,
                sampler=sampler,
                sample_params={"tune": 2, "draws": 2, "chains": 4, "njobs": 8},
                vi_params={
                    "n": 20000,
                    "method": "advi",
                    "callbacks": [
                        pm.callbacks.CheckParametersConvergence(
                            diff="absolute", tolerance=0.01, every=50
                        )
                    ],
                    "draws": 500,
                },
                **kwargs,
            )
            self.threshold = 0.5

    def _fit(
        self,
        X,
        Y,
        sampler="variational",
        tune=500,
        draws=500,
        vi_params={
            "n": 20000,
            "method": "advi",
            "callbacks": [CheckParametersConvergence()],
        },
        **kwargs,
    ):
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        self.construct_model(X, Y)
        fit_pymc3_model(self, sampler, draws, tune, vi_params, **kwargs)

    def _predict_scores_fixed(self, X, **kwargs):
        d = dict(pm.summary(self.trace)["mean"])
        intercept = 0.0
        weights = np.array(
            [d["weights[{}]".format(i)] for i in range(self.n_object_features_fit_)]
        )
        if "intercept" in d:
            intercept = intercept + d["intercept"]
        return np.dot(X, weights) + intercept

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ChoiceFunctions.predict_for_scores(self, scores, **kwargs)

    def set_tunable_parameters(self, regularization="l1", **point):
        """
            Set tunable parameters of the Generalized Linear model to the values provided.

            Parameters
            ----------
            regularization : {‘l1’, ‘l2’}, string
               Regularizer function (L1 or L2) applied to the `kernel` weights matrix
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        self.regularization = regularization
        self.model = None
        self.trace = None
        self.trace_vi = None
        self.Xt = None
        self.Yt = None
        self.p = None
        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support"
                " tunable parameters called: {}".format(print_dictionary(point))
            )



import logging

from sklearn.model_selection import train_test_split
from sklearn.utils import check_random_state

from csrank.dataset_reader.util import standardize_features
from .dataset_reader import DatasetReader


class SyntheticDatasetGenerator(DatasetReader):
    def __init__(
        self,
        learning_problem,
        n_train_instances=10000,
        n_test_instances=10000,
        random_state=None,
        standardize=True,
        **kwargs,
    ):
        super(SyntheticDatasetGenerator, self).__init__(
            learning_problem=learning_problem, dataset_folder=None, **kwargs
        )
        self.random_state = check_random_state(random_state)
        self.dataset_function = None
        self.kwargs = kwargs
        self.n_test_instances = n_test_instances
        self.n_train_instances = n_train_instances
        self.logger = logging.getLogger(SyntheticDatasetGenerator.__name__)
        self.standardize = standardize
        self.logger.info("Key word arguments {}".format(kwargs))

    def __load_dataset__(self):
        pass

    def splitter(self, iter):
        for i in iter:
            seed = self.random_state.randint(2 ** 32, dtype="uint32")
            total_instances = self.n_test_instances + self.n_train_instances
            self.kwargs["n_instances"] = total_instances
            X, Y = self.dataset_function(**self.kwargs, seed=seed)
            x_train, x_test, y_train, y_test = train_test_split(
                X, Y, random_state=self.random_state, test_size=self.n_test_instances
            )
            if self.standardize:
                x_train, x_test = standardize_features(x_train, x_test)
            yield x_train, y_train, x_test, y_test

    def get_dataset_dictionaries(self, lengths=[5, 6]):
        x_train = dict()
        y_train = dict()
        x_test = dict()
        y_test = dict()
        for n_obj in lengths:
            self.kwargs["n_objects"] = n_obj

            seed = self.random_state.randint(2 ** 32, dtype="uint32")
            total_instances = self.n_test_instances + self.n_train_instances
            self.kwargs["n_instances"] = total_instances
            X, Y = self.dataset_function(**self.kwargs, seed=seed)
            x_1, x_2, y_1, y_2 = train_test_split(
                X, Y, random_state=self.random_state, test_size=self.n_test_instances
            )
            if self.standardize:
                x_1, x_2 = standardize_features(x_1, x_2)
            x_train[n_obj], x_test[n_obj], y_train[n_obj], y_test[n_obj] = (
                x_1,
                x_2,
                y_1,
                y_2,
            )
        self.logger.info("Done")
        return x_train, y_train, x_test, y_test

    def get_single_train_test_split(self):
        seed = self.random_state.randint(2 ** 32, dtype="uint32")
        total_instances = self.n_test_instances + self.n_train_instances
        self.kwargs["n_instances"] = total_instances
        self.X, self.Y = self.dataset_function(**self.kwargs, seed=seed)
        self.__check_dataset_validity__()
        x_train, x_test, y_train, y_test = train_test_split(
            self.X,
            self.Y,
            random_state=self.random_state,
            test_size=self.n_test_instances,
        )
        if self.standardize:
            x_train, x_test = standardize_features(x_train, x_test)
        self.logger.info("Done")

        return x_train, y_train, x_test, y_test


class SyntheticIterator(object):
    def __init__(self, dataset_function, **params):
        """
        Infinite iterator over a synthetic dataset generator.

        Parameters
        ----------
        dataset_function : callable
            Returns a tuple (inputs, targets) when called
        params : dict
            Parameters to be passed to `dataset_function` when called
        """
        self.params = params
        self.func = dataset_function

    def __iter__(self):
        return self

    def __next__(self):
        return self.func(**self.params)

    def __len__(self):
        """Return a constant to allow for steps per epoch."""
        return 100



from .baseline import RandomBaselineRanker
from .cmp_net import CmpNet
from .expected_rank_regression import ExpectedRankRegression
from .fate_object_ranker import FATEObjectRanker
from .fatelinear_object_ranker import FATELinearObjectRanker
from .feta_object_ranker import FETAObjectRanker
from .fetalinear_object_ranker import FETALinearObjectRanker
from .list_net import ListNet
from .rank_net import RankNet
from .rank_svm import RankSVM

__all__ = [
    "CmpNet",
    "ExpectedRankRegression",
    "FATEObjectRanker",
    "FATELinearObjectRanker",
    "FETAObjectRanker",
    "FETALinearObjectRanker",
    "ListNet",
    "RankNet",
    "RankSVM",
    "RandomBaselineRanker",
]



from csrank.constants import OBJECT_RANKING
from csrank.numpy_util import scores_to_rankings
from ..tag_genome_reader import TagGenomeDatasetReader


class TagGenomeObjectRankingDatasetReader(TagGenomeDatasetReader):
    def __init__(self, dataset_type="similarity", **kwargs):
        super(TagGenomeObjectRankingDatasetReader, self).__init__(
            learning_problem=OBJECT_RANKING, **kwargs
        )
        dataset_func_dict = {
            "nearest_neighbour": self.make_nearest_neighbour_dataset,
            "critique_fit_less": self.make_critique_fit_dataset(direction=-1),
            "critique_fit_more": self.make_critique_fit_dataset(direction=1),
        }
        if dataset_type not in dataset_func_dict:
            raise ValueError(
                f"dataset_type must be one of {set(dataset_func_dict.keys())}"
            )
        self.logger.info("Dataset type: {}".format(dataset_type))
        self.dataset_function = dataset_func_dict[dataset_type]

    def make_nearest_neighbour_dataset(self, n_instances, n_objects, seed, **kwargs):
        X, scores = super().make_nearest_neighbour_dataset(
            n_instances=n_instances, n_objects=n_objects, seed=seed
        )
        # Higher the similarity lower the rank of the object
        Y = scores_to_rankings(scores)
        return X, Y

    def make_critique_fit_dataset(self, direction):
        def dataset_generator(n_instances, n_objects, seed, **kwargs):
            X, scores = super(
                TagGenomeObjectRankingDatasetReader, self
            ).make_critique_fit_dataset(
                n_instances=n_instances,
                n_objects=n_objects,
                seed=seed,
                direction=direction,
            )
            Y = scores_to_rankings(scores)
            return X, Y

        return dataset_generator



from abc import ABCMeta
from abc import abstractmethod


class Tunable(metaclass=ABCMeta):
    @abstractmethod
    def set_tunable_parameters(self, **point):
        """ Set tunable parameters of the algorithm to the values provided.

        Parameters
        ----------
        point : dict
            Dictionary containing parameter values

        """
        raise NotImplementedError

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Tunable:
            has_set_tunable_param = any(
                "set_tunable_parameters" in B.__dict__ for B in C.__mro__
            )
            if has_set_tunable_param:
                return True
        return NotImplemented



import glob
import logging
import os

import numpy as np
from sklearn.model_selection import ShuffleSplit
from sklearn.utils import check_random_state

from csrank.constants import LABEL_RANKING
from csrank.numpy_util import ranking_ordering_conversion
from ..dataset_reader import DatasetReader

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")


class IntelligentSystemGroupDatasetReader(DatasetReader):
    def __init__(self, random_state=None, **kwargs):
        super(IntelligentSystemGroupDatasetReader, self).__init__(
            learning_problem=LABEL_RANKING,
            dataset_folder="intelligent_system_data",
            **kwargs,
        )

        self.logger = logging.getLogger(IntelligentSystemGroupDatasetReader.__name__)
        self.train_files = glob.glob(os.path.join(self.dirname, "*.txt"))
        self.train_files.sort()
        self.train_files_names = []
        self.random_state = check_random_state(random_state)
        self.X = dict()
        self.Y = dict()
        self.__load_dataset__()

    def __load_dataset__(self):
        for i, file in enumerate(self.train_files):
            raw_data = pd.read_csv(file, skiprows=[1], delimiter="\t")
            feature_cols = [x for x in raw_data.columns if x[0] == "A"]
            label_cols = [x for x in raw_data.columns if x[0] == "L" or "O"]
            features = raw_data.loc[:, feature_cols]
            labels = raw_data.loc[:, label_cols]
            if "O1" in raw_data.columns:
                labels = ranking_ordering_conversion(labels)
            labels -= np.min(labels)
            name = os.path.basename(file).split(".")[0]
            self.train_files_names.append(name)
            self.X[name] = features.as_matrix()
            self.Y[name] = labels.as_matrix()
        self.logger.info("Dataset files: " + repr(self.train_files_names))

    def get_single_train_test_split(self, name="cold"):
        cv_iter = ShuffleSplit(
            n_splits=1, test_size=0.3, random_state=self.random_state
        )
        (train_idx, test_idx) = list(cv_iter.split(self.X[name]))[0]
        return (
            self.X[name][train_idx],
            self.Y[name][train_idx],
            self.X[name][test_idx],
            self.Y[name][test_idx],
        )

    def get_dataset_dictionaries(self, name="cold"):
        return self.X[name], self.Y[name]



import os

from keras.optimizers import SGD
import numpy as np
import pytest
import tensorflow as tf

from csrank import PairedCombinatorialLogit
from csrank.constants import CMPNET
from csrank.constants import ERR
from csrank.constants import FATE_RANKER
from csrank.constants import FATELINEAR_RANKER
from csrank.constants import FETA_RANKER
from csrank.constants import FETALINEAR_RANKER
from csrank.constants import LISTNET
from csrank.constants import RANKNET
from csrank.constants import RANKSVM
from csrank.metrics_np import zero_one_accuracy_np
from csrank.metrics_np import zero_one_rank_loss_for_scores_ties_np
from csrank.objectranking import *
from csrank.objectranking.fate_object_ranker import FATEObjectRanker

optimizer = SGD(lr=1e-3, momentum=0.9, nesterov=True)

object_rankers = {
    FATELINEAR_RANKER: (
        FATELinearObjectRanker,
        {"n_hidden_set_units": 12, "batch_size": 1},
        (0.0, 1.0),
    ),
    FETALINEAR_RANKER: (FETALinearObjectRanker, {}, (0.0, 1.0)),
    FETA_RANKER: (
        FETAObjectRanker,
        {"add_zeroth_order_model": True, "optimizer": optimizer},
        (0.0, 1.0),
    ),
    RANKNET: (RankNet, {"optimizer": optimizer}, (0.0, 1.0)),
    CMPNET: (CmpNet, {"optimizer": optimizer}, (0.0, 1.0)),
    LISTNET: (ListNet, {"n_top": 3, "optimizer": optimizer}, (0.0, 1.0)),
    ERR: (ExpectedRankRegression, {}, (0.0, 1.0)),
    RANKSVM: (RankSVM, {}, (0.0, 1.0)),
    FATE_RANKER: (
        FATEObjectRanker,
        {
            "n_hidden_joint_layers": 1,
            "n_hidden_set_layers": 1,
            "n_hidden_joint_units": 5,
            "n_hidden_set_units": 5,
            "optimizer": optimizer,
        },
        (0.0, 1.0),
    ),
}


@pytest.fixture(scope="module")
def trivial_ranking_problem():
    random_state = np.random.RandomState(123)
    x = random_state.randn(2, 5, 1)
    y_true = x.argsort(axis=1).argsort(axis=1).squeeze(axis=-1)
    return x, y_true


def check_params_tunable(tunable_obj, params, rtol=1e-2, atol=1e-4):
    for key, value in params.items():
        if hasattr(tunable_obj, key):
            expected = getattr(tunable_obj, key)
            if isinstance(value, int) or isinstance(value, float):
                if (
                    isinstance(tunable_obj, PairedCombinatorialLogit)
                    and key == "n_nests"
                ):
                    tunable_obj.n_nests == tunable_obj.n_objects_fit_ * (
                        tunable_obj.n_objects_fit_ - 1
                    ) / 2
                else:
                    assert np.isclose(
                        expected, value, rtol=rtol, atol=atol, equal_nan=False
                    )
            else:
                assert value == expected
        elif key == "learning_rate" and hasattr(tunable_obj, "optimizer"):
            key = (
                "learning_rate"
                if "learning_rate" in tunable_obj.optimizer.get_config()
                else "lr"
            )
            learning_rate = tunable_obj.optimizer.get_config().get(key, 0.0)
            assert np.isclose(
                learning_rate, value, rtol=rtol, atol=atol, equal_nan=False
            )
        elif key == "reg_strength" and hasattr(tunable_obj, "kernel_regularizer"):
            config = tunable_obj.kernel_regularizer.get_config()
            val1 = np.isclose(
                config["l1"], value, rtol=rtol, atol=atol, equal_nan=False
            )
            val2 = np.isclose(
                config["l2"], value, rtol=rtol, atol=atol, equal_nan=False
            )
            assert val1 or val2


@pytest.mark.parametrize("ranker_name", list(object_rankers.keys()))
def test_object_ranker_fixed(trivial_ranking_problem, ranker_name):
    tf.set_random_seed(0)
    os.environ["KERAS_BACKEND"] = "tensorflow"
    np.random.seed(123)
    x, y = trivial_ranking_problem
    ranker, params, (loss, acc) = object_rankers[ranker_name]
    ranker = ranker(**params)
    if "linear" in ranker_name:
        ranker.fit(x, y, epochs=10, validation_split=0, verbose=False)
    else:
        ranker.fit(x, y, epochs=100, validation_split=0, verbose=False)
    pred_scores = ranker.predict_scores(x)
    pred_loss = zero_one_rank_loss_for_scores_ties_np(y, pred_scores)
    rtol = 1e-2
    atol = 1e-4
    assert np.isclose(loss, pred_loss, rtol=rtol, atol=atol, equal_nan=False)
    pred = ranker.predict_for_scores(pred_scores)
    pred_2 = ranker.predict(x)
    pred_acc = zero_one_accuracy_np(pred, pred_2)
    assert np.isclose(1.0, pred_acc, rtol=rtol, atol=atol, equal_nan=False)
    pred_acc = zero_one_accuracy_np(pred, y)
    assert np.isclose(acc, pred_acc, rtol=rtol, atol=atol, equal_nan=False)
    params = {
        "n_hidden": 20,
        "n_units": 20,
        "n_hidden_set_units": 2,
        "n_hidden_set_layers": 10,
        "n_hidden_joint_units": 2,
        "n_hidden_joint_layers": 10,
        "reg_strength": 1e-3,
        "learning_rate": 1e-1,
        "batch_size": 32,
        "alpha": 0.5,
        "l1_ratio": 0.7,
        "tol": 1e-2,
        "C": 10,
        "n_mixtures": 10,
        "n_nests": 5,
        "regularization": "l2",
    }
    ranker.set_tunable_parameters(**params)
    check_params_tunable(ranker, params, rtol, atol)



import numpy as np
from sklearn.preprocessing import LabelBinarizer


def sub_sampling_discrete_choices_from_scores(Xt, Yt, n_objects=5):
    bucket_size = int(Xt.shape[1] / n_objects)
    X_train = []
    Y_train = []
    for i in range(bucket_size):
        X = np.copy(Xt)
        Y = np.copy(Yt)
        rs = np.random.RandomState(42 + i)
        idx = rs.randint(bucket_size, size=(len(X), n_objects))
        idx += np.arange(start=0, stop=X.shape[1], step=bucket_size)[:n_objects]
        X = X[np.arange(X.shape[0])[:, None], idx]
        Y = Y[np.arange(X.shape[0])[:, None], idx]
        Y = Y.argmax(axis=1)
        if len(X_train) == 0:
            X_train = X
            Y_train = Y
        else:
            Y_train = np.concatenate([Y_train, Y], axis=0)
            X_train = np.concatenate([X_train, X], axis=0)
    if len(Y_train) != 0:
        Y_train = convert_to_label_encoding(Y_train, n_objects)
    return X_train, Y_train


def sub_sampling_discrete_choices_from_relevance(Xt, Yt, n_objects=5):
    X_train = []
    Y_train = []
    bucket_size = int(Xt.shape[1] / n_objects)
    rs = np.random.RandomState(42 + n_objects)
    for x, y in zip(Xt, Yt):
        choices = np.append(np.where(y == 1)[0], np.where(y == 2)[0])
        zeros = np.delete(np.arange(0, Xt.shape[1]), choices)
        if len(zeros) != 0:
            if len(choices) > bucket_size:
                idx = choices
            else:
                idx = rs.choice(choices, size=bucket_size)
            if len(zeros) > n_objects - 1:
                idys = np.array(
                    [
                        rs.choice(zeros, size=(n_objects - 1), replace=False)
                        for i in range(len(idx))
                    ]
                )
            else:
                idys = rs.choice(zeros, size=(len(idx), (n_objects - 1)), replace=True)
            idx = np.append(idys, idx[:, None], axis=1)
            for i in idx:
                np.random.shuffle(i)
            if len(X_train) == 0:
                X_train = x[idx]
                Y_train = np.argmax(y[idx], axis=1)
            else:
                Y_train = np.concatenate([Y_train, np.argmax(y[idx], axis=1)])
                X_train = np.concatenate([X_train, x[idx]], axis=0)
    if len(Y_train) != 0:
        Y_train = convert_to_label_encoding(Y_train, n_objects)
    return X_train, Y_train


def convert_to_label_encoding(Y, n_objects):
    lb = LabelBinarizer().fit(np.arange(n_objects))
    return lb.transform(Y)


def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    return vector / np.linalg.norm(vector)


def angle_between(v1, v2):
    """ Returns the angle in radians between vectors 'v1' and 'v2'::

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))



from itertools import permutations
import logging

from keras import backend as K
from keras import Input
from keras import Model
from keras import optimizers
from keras.layers import concatenate
from keras.layers import Dense
from keras.optimizers import SGD
from keras.regularizers import l2
import numpy as np
from sklearn.utils import check_random_state
import tensorflow as tf

from csrank.constants import allowed_dense_kwargs
from csrank.layers import NormalizedDense
from csrank.learner import Learner
from csrank.util import print_dictionary


class CmpNetCore(Learner):
    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        loss_function="binary_crossentropy",
        batch_normalization=True,
        kernel_regularizer=l2(1e-4),
        kernel_initializer="lecun_normal",
        activation="relu",
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        metrics=["binary_accuracy"],
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        self.logger = logging.getLogger("CmpNet")
        self.batch_normalization = batch_normalization
        self.activation = activation
        self.hash_file = None

        self.batch_size = batch_size

        self.metrics = metrics
        self.kernel_regularizer = kernel_regularizer
        self.kernel_initializer = kernel_initializer
        self.loss_function = loss_function

        self.optimizer = optimizers.get(optimizer)
        self._optimizer_config = self.optimizer.get_config()

        self.n_hidden = n_hidden
        self.n_units = n_units
        keys = list(kwargs.keys())
        for key in keys:
            if key not in allowed_dense_kwargs:
                del kwargs[key]
        self.kwargs = kwargs
        self.threshold_instances = int(1e10)
        self.random_state = random_state
        self.model = None

    def _construct_layers(self, **kwargs):

        self.output_node = Dense(
            1, activation="sigmoid", kernel_regularizer=self.kernel_regularizer
        )

        self.x1 = Input(shape=(self.n_object_features_fit_,))
        self.x2 = Input(shape=(self.n_object_features_fit_,))
        if self.batch_normalization:
            self.hidden_layers = [
                NormalizedDense(self.n_units, name="hidden_{}".format(x), **kwargs)
                for x in range(self.n_hidden)
            ]
        else:
            self.hidden_layers = [
                Dense(self.n_units, name="hidden_{}".format(x), **kwargs)
                for x in range(self.n_hidden)
            ]
        assert len(self.hidden_layers) == self.n_hidden

    def _convert_instances_(self, X, Y):
        raise NotImplementedError

    def construct_model(self):
        """
            Construct the CmpNet which is used to approximate the :math:`U_1(x_i,x_j)`. For each pair of objects in
            :math:`x_i, x_j \\in Q` we construct two sub-networks with weight sharing in all hidden layers.
            The output of these networks are connected to two sigmoid units that produces the outputs of the network,
            i.e., :math:`U(x_1,x_2), U(x_2,x_1)` for each pair of objects are evaluated. :math:`U(x_1,x_2)` is a measure
            of how favorable it is to choose :math:`x_1` over :math:`x_2`.

            Returns
            -------
            model: keras :class:`Model`
                Neural network to learn the CmpNet utility score
        """
        x1x2 = concatenate([self.x1, self.x2])
        x2x1 = concatenate([self.x2, self.x1])
        self.logger.debug("Creating the model")
        for hidden in self.hidden_layers:
            x1x2 = hidden(x1x2)
            x2x1 = hidden(x2x1)
        merged_left = concatenate([x1x2, x2x1])
        merged_right = concatenate([x2x1, x1x2])
        N_g = self.output_node(merged_left)
        N_l = self.output_node(merged_right)
        merged_output = concatenate([N_g, N_l])
        model = Model(inputs=[self.x1, self.x2], outputs=merged_output)
        model.compile(
            loss=self.loss_function, optimizer=self.optimizer, metrics=self.metrics
        )
        return model

    def fit(
        self, X, Y, epochs=10, callbacks=None, validation_split=0.1, verbose=0, **kwd
    ):
        """
            Fit a generic preference learning CmptNet on the provided set of queries X and preferences Y of those
            objects. The provided queries and corresponding preferences are of a fixed size (numpy arrays).
            For learning this network the binary cross entropy loss function for a pair of objects
            :math:`x_i, x_j \\in Q` is defined as:

            .. math::

                C_{ij} =  -\\tilde{P_{ij}}(0)\\cdot \\log(U(x_i,x_j)) - \\tilde{P_{ij}}(1) \\cdot \\log(U(x_j,x_i)) \\ ,

            where :math:`\\tilde{P_{ij}}` is ground truth probability of the preference of :math:`x_i` over :math:`x_j`.
            :math:`\\tilde{P_{ij}} = (1,0)` if :math:`x_i \\succ x_j` else :math:`\\tilde{P_{ij}} = (0,1)`.

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Preferences in form of Orderings or Choices for given n_objects
            epochs : int
                Number of epochs to run if training for a fixed query size
            callbacks : list
                List of callbacks to be called during optimization
            validation_split : float (range : [0,1])
                Percentage of instances to split off to validate on
            verbose : bool
                Print verbose information
            **kwd :
                Keyword arguments for the fit function
        """
        self.random_state_ = check_random_state(self.random_state)
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        x1, x2, y_double = self._convert_instances_(X, Y)

        self.logger.debug("Instances created {}".format(x1.shape[0]))
        self._construct_layers(
            kernel_regularizer=self.kernel_regularizer,
            kernel_initializer=self.kernel_initializer,
            activation=self.activation,
            **self.kwargs,
        )
        self.model = self.construct_model()

        self.logger.debug("Finished Creating the model, now fitting started")
        self.model.fit(
            [x1, x2],
            y_double,
            batch_size=self.batch_size,
            epochs=epochs,
            callbacks=callbacks,
            validation_split=validation_split,
            verbose=verbose,
            **kwd,
        )
        self.logger.debug("Fitting Complete")

    def predict_pair(self, a, b, **kwargs):
        return self.model.predict([a, b], **kwargs)

    def _predict_scores_fixed(self, X, **kwargs):
        n_instances, n_objects, n_features = X.shape
        self.logger.info(
            "Test Set instances {} objects {} features {}".format(*X.shape)
        )
        n2 = n_objects * (n_objects - 1)
        pairs = np.empty((n2, 2, n_features))
        scores = np.empty((n_instances, n_objects))
        for n in range(n_instances):
            for k, (i, j) in enumerate(permutations(range(n_objects), 2)):
                pairs[k] = (X[n, i], X[n, j])
            result = self.predict_pair(pairs[:, 0], pairs[:, 1], **kwargs)[:, 0]
            scores[n] = result.reshape(n_objects, n_objects - 1).mean(axis=1)
            del result
        del pairs
        self.logger.info("Done predicting scores")

        return scores

    def clear_memory(self, **kwargs):
        """
            Clear the memory, restores the currently fitted model back to prevent memory leaks.

            Parameters
            ----------
            **kwargs :
                Keyword arguments for the function
        """
        if self.hash_file is not None:
            self.model.save_weights(self.hash_file)
            K.clear_session()
            sess = tf.Session()
            K.set_session(sess)

            self.optimizer = self.optimizer.from_config(self._optimizer_config)
            self._construct_layers(
                kernel_regularizer=self.kernel_regularizer,
                kernel_initializer=self.kernel_initializer,
                activation=self.activation,
                **self.kwargs,
            )
            self.model = self.construct_model()
            self.model.load_weights(self.hash_file)
        else:
            self.logger.info("Cannot clear the memory")

    def set_tunable_parameters(
        self,
        n_hidden=32,
        n_units=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        """
            Set tunable parameters of the CmpNet network to the values provided.

            Parameters
            ----------
            n_hidden: int
                Number of hidden layers used in the scoring network
            n_units: int
                Number of hidden units in each layer of the scoring network
            reg_strength: float
                Regularization strength of the regularizer function applied to the `kernel` weights matrix
            learning_rate: float
                Learning rate of the stochastic gradient descent algorithm used by the network
            batch_size: int
                Batch size to use during training
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        self.n_hidden = n_hidden
        self.n_units = n_units
        self.kernel_regularizer = l2(reg_strength)
        self.batch_size = batch_size
        self.optimizer = self.optimizer.from_config(self._optimizer_config)
        K.set_value(self.optimizer.lr, learning_rate)
        self._construct_layers(
            kernel_regularizer=self.kernel_regularizer,
            kernel_initializer=self.kernel_initializer,
            activation=self.activation,
            **self.kwargs,
        )
        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support"
                " tunable parameters called: {}".format(print_dictionary(point))
            )



from abc import ABCMeta

import numpy as np

from csrank.constants import CHOICE_FUNCTION
from csrank.metrics_np import f1_measure
from csrank.util import progress_bar

__all__ = ["ChoiceFunctions"]


class ChoiceFunctions(metaclass=ABCMeta):
    @property
    def learning_problem(self):
        return CHOICE_FUNCTION

    def predict_for_scores(self, scores, **kwargs):
        """
            Binary choice vector :math:`y` represents the choices amongst the objects in :math:`Q`, such that
            :math:`y(k) = 1` represents that the object :math:`x_k` is chosen and :math:`y(k) = 0` represents it is not
            chosen. Predict choices for the scores for a given collection of sets of objects (query sets).

            Parameters
            ----------
            scores : dict or numpy array
                Dictionary with a mapping from query set size to numpy arrays
                or a single numpy array of size containing scores of each object of size:
                (n_instances, n_objects)


            Returns
            -------
            Y : dict or numpy array
                Dictionary with a mapping from query set size to numpy arrays
                or a single numpy array containing predicted choice vectors of size:
                (n_instances, n_objects)
        """

        if isinstance(scores, dict):
            result = dict()
            for n, score in scores.items():
                result[n] = score > self.threshold
                result[n] = np.array(result[n], dtype=int)
        else:
            result = scores > self.threshold
            result = np.array(result, dtype=int)
        return result

    def _tune_threshold(self, X_val, Y_val, thin_thresholds=1, verbose=0):
        scores = self.predict_scores(X_val)
        probabilities = np.unique(scores)[::thin_thresholds]
        threshold = 0.0
        best = f1_measure(Y_val, scores > threshold)
        try:
            for i, p in enumerate(probabilities):
                pred = scores > p
                f1 = f1_measure(Y_val, pred)
                if f1 > best:
                    threshold = p
                    best = f1
                if verbose == 1:
                    progress_bar(i, len(probabilities), status="Tuning threshold")
        except KeyboardInterrupt:
            self.logger.info("Keyboard interrupted")
        self.logger.info(
            "Tuned threshold, obtained {:.2f} which achieved"
            " a micro F1-measure of {:.2f}".format(threshold, best)
        )
        return threshold



import math
import os

from keras import Sequential
from keras.layers import Dense
from keras.optimizers import SGD
import numpy as np
import pytest
import tensorflow as tf

from csrank.callbacks import EarlyStoppingWithWeights
from csrank.callbacks import LRScheduler
from csrank.tests.test_ranking import check_params_tunable

callbacks_dict = {
    "EarlyStopping": (EarlyStoppingWithWeights, {"patience": 5, "min_delta": 5e-2}),
    "LRScheduler": (LRScheduler, {"epochs_drop": 5, "drop": 0.9}),
}


@pytest.fixture(scope="module")
def trivial_classification_problem():
    random_state = np.random.RandomState(123)
    x = random_state.randn(200, 2)
    w = random_state.rand(2)
    y = 1.0 / (1.0 + np.exp(-np.dot(x, w)))
    y_true = np.array(y > 0.5, dtype=np.int64)
    return x, y_true


def create_model():
    lr = 0.015
    model = Sequential()
    model.add(Dense(10, activation="relu"))
    model.add(Dense(5, activation="relu"))
    model.add(Dense(1, activation="sigmoid"))
    model.compile(optimizer=SGD(lr=lr), loss="binary_crossentropy")
    return model, lr


@pytest.mark.parametrize("name", list(callbacks_dict.keys()))
def test_callbacks(trivial_classification_problem, name):
    tf.set_random_seed(0)
    os.environ["KERAS_BACKEND"] = "tensorflow"
    np.random.seed(123)
    x, y = trivial_classification_problem
    epochs = 15
    model, init_lr = create_model()
    callback, params = callbacks_dict[name]
    callback = callback(**params)
    callbacks = [callback]
    model.fit(x, y, epochs=epochs, callbacks=callbacks, validation_split=0.1)
    rtol = 1e-2
    atol = 5e-4
    if name == "LRScheduler":
        epochs_drop, drop = params["epochs_drop"], params["drop"]
        step = math.floor(epochs / epochs_drop)
        actual_lr = init_lr * math.pow(drop, step)
        key = (
            "learning_rate" if "learning_rate" in model.optimizer.get_config() else "lr"
        )
        learning_rate = model.optimizer.get_config().get(key, 0.0)
        assert np.isclose(
            actual_lr, learning_rate, rtol=rtol, atol=atol, equal_nan=False
        )
    elif name == "EarlyStopping":
        assert callback.stopped_epoch == 6
    params = {"epochs_drop": 100, "drop": 0.5, "patience": 10, "min_delta": 1e-5}
    callback.set_tunable_parameters(**params)
    check_params_tunable(callback, params, rtol, atol)



import logging

from sklearn.utils import check_random_state

from csrank.learner import Learner
from .discrete_choice import DiscreteObjectChooser


class RandomBaselineDC(DiscreteObjectChooser, Learner):
    def __init__(self, random_state=None, **kwargs):
        """
            Baseline assigns the average number of chosen objects in the given choice sets and chooses all the objects.

            :param kwargs: Keyword arguments for the algorithms
        """

        self.logger = logging.getLogger(RandomBaselineDC.__name__)
        self.random_state = random_state
        self.model = None

    def fit(self, X, Y, **kwd):
        self.random_state_ = check_random_state(self.random_state)

    def _predict_scores_fixed(self, X, **kwargs):
        n_instances, n_objects, n_features = X.shape
        return self.random_state_.rand(n_instances, n_objects)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return DiscreteObjectChooser.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def set_tunable_parameters(self, **point):
        pass



import numpy as np
from sklearn.utils import check_random_state

from csrank.constants import CHOICE_FUNCTION
from ..synthetic_dataset_generator import SyntheticDatasetGenerator


class ChoiceDatasetGenerator(SyntheticDatasetGenerator):
    def __init__(self, dataset_type="pareto", **kwargs):
        super(ChoiceDatasetGenerator, self).__init__(
            learning_problem=CHOICE_FUNCTION, **kwargs
        )
        dataset_function_options = {
            "linear": self.make_latent_linear_choices,
            "pareto": self.make_globular_pareto_choices,
        }
        if dataset_type not in dataset_function_options:
            raise ValueError(
                f"dataset_type must be one of {set(dataset_function_options.keys())}"
            )
        self.dataset_function = dataset_function_options[dataset_type]

    def get_single_train_test_split(self):
        return super(ChoiceDatasetGenerator, self).get_single_train_test_split()

    def get_train_test_datasets(self, n_datasets=5):
        return super(ChoiceDatasetGenerator, self).get_train_test_datasets(
            n_datasets=n_datasets
        )

    def make_globular_pareto_choices(
        self,
        n_instances=10000,
        n_features=2,
        n_objects=10,
        seed=42,
        cluster_spread=1.0,
        cluster_size=10,
        **kwargs,
    ):
        def pareto_front(X, signs=None):
            n_points, n_attributes = X.shape
            if signs is None:
                signs = -np.ones(n_attributes)
            pareto = np.ones(n_points, dtype=bool)
            for i, attr in enumerate(X):
                pareto[i] = np.all(
                    np.any((X * signs[None, :]) <= (attr * signs), axis=1)
                )
            return pareto

        def sample_unit_ball(n_inst=10000, n_features=2, rng=None, radius=1.0):
            rng = check_random_state(rng)
            X = rng.randn(n_inst, n_features)
            u = rng.uniform(size=n_inst)[:, None]
            X /= np.linalg.norm(X, axis=1, ord=2)[:, None]
            X *= radius * u
            return X

        def make_randn_pareto_choices(
            n_instances=10000, n_features=2, n_objects=10, data_seed=None, center=0.0
        ):
            """Generate random objects from a d-dimensional isometric normal distribution.

            This should be the easiest possible Pareto-problem, since the model can learn
            a latent-utility which scores how likely a point is on the front (independent
            of the other points)."""
            rand = check_random_state(data_seed)
            X = rand.randn(n_instances, n_objects, n_features)
            Y = np.empty((n_instances, n_objects), dtype=bool)
            for i in range(n_instances):
                Y[i] = pareto_front(X[i])
            return X + center, Y

        rand = check_random_state(seed)
        X = np.empty((n_instances, n_objects, n_features))
        Y = np.empty((n_instances, n_objects), dtype=int)
        for i in range(int(n_instances / cluster_size)):
            center = sample_unit_ball(
                n_inst=1, n_features=n_features, rng=rand, radius=cluster_spread
            )
            x, y = make_randn_pareto_choices(
                n_instances=cluster_size,
                n_features=n_features,
                n_objects=n_objects,
                data_seed=rand,
                center=center,
            )
            X[i * cluster_size : (i + 1) * cluster_size] = x
            Y[i * cluster_size : (i + 1) * cluster_size] = y
        return X, Y

    def make_latent_linear_choices(
        self,
        n_instances=10000,
        n_features=2,
        n_objects=6,
        n_rep_units=5,
        threshold=0.0,
        seed=42,
        **kwargs,
    ):
        rand = check_random_state(seed)
        ranw = check_random_state(rand.randint(2 ** 32, dtype="uint32"))
        X = rand.uniform(-1, 1, size=(n_instances, n_objects, n_features))
        W_rep = ranw.randn(n_features, n_rep_units)
        rep = X.dot(W_rep).mean(axis=-2)
        w_join = ranw.randn(n_features + n_rep_units)
        joint_matrix = np.empty(
            (n_objects, n_instances, n_features + n_rep_units), dtype=np.float32
        )
        for i in range(n_objects):
            joint_matrix[i] = np.concatenate((X[:, i], rep), axis=-1)
        scores = joint_matrix.dot(w_join)
        Y = scores > threshold
        Y = Y.astype(int)
        return X, Y.T
